(* *********************************************************************)
(*                                                                     *)
(*              The Compcert verified compiler                         *)
(*                                                                     *)
(*          Xavier Leroy, INRIA Paris-Rocquencourt                     *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique.  All rights reserved.  This file is distributed       *)
(*  under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 2 of the License, or  *)
(*  (at your option) any later version.  This file is also distributed *)
(*  under the terms of the INRIA Non-Commercial License Agreement.     *)
(*                                                                     *)
(* *********************************************************************)

(** The Clight language: a simplified version of Compcert C where all
  expressions are pure and assignments and function calls are
  statements, not expressions. *)

Require Import Coqlib.
Require Import Errors.
Require Import Maps.
Require Import Integers.
Require Import Floats.
Require Import Values.
Require Import AST.
Require Import Memory.
Require Import Events.
Require Import Globalenvs.
Require Import Smallstep.
Require Import Ctypes.
Require Import Cop.
Require Import StrongInject.

Section WITHMEM.
Context `{M: Mem.MEM}.

(** * Abstract syntax *)

(** ** Expressions *)

(** Clight expressions correspond to the "pure" subset of C expressions.
  The main omissions are string literals and assignment operators
  ([=], [+=], [++], etc).  In Clight, assignment is a statement,
  not an expression.  Additionally, an expression can also refer to
  temporary variables, which are a separate class of local variables
  that do not reside in memory and whose address cannot be taken.

  As in Compcert C, all expressions are annotated with their types,
  as needed to resolve operator overloading and type-dependent behaviors. *)

Inductive expr : Type :=
  | Econst_int: int -> type -> expr       (**r integer literal *)
  | Econst_float: float -> type -> expr   (**r float literal *)
  | Evar: ident -> type -> expr           (**r variable *)
  | Etempvar: ident -> type -> expr       (**r temporary variable *)
  | Ederef: expr -> type -> expr          (**r pointer dereference (unary [*]) *)
  | Eaddrof: expr -> type -> expr         (**r address-of operator ([&]) *)
  | Eunop: unary_operation -> expr -> type -> expr  (**r unary operation *)
  | Ebinop: binary_operation -> expr -> expr -> type -> expr (**r binary operation *)
  | Ecast: expr -> type -> expr   (**r type cast ([(ty) e]) *)
  | Efield: expr -> ident -> type -> expr. (**r access to a member of a struct or union *)

(** [sizeof] and [alignof] are derived forms. *)

Definition Esizeof (ty' ty: type) : expr := Econst_int (Int.repr(sizeof ty')) ty.
Definition Ealignof (ty' ty: type) : expr := Econst_int (Int.repr(alignof ty')) ty.

(** Extract the type part of a type-annotated Clight expression. *)

Definition typeof (e: expr) : type :=
  match e with
  | Econst_int _ ty => ty
  | Econst_float _ ty => ty
  | Evar _ ty => ty
  | Etempvar _ ty => ty
  | Ederef _ ty => ty
  | Eaddrof _ ty => ty
  | Eunop _ _ ty => ty
  | Ebinop _ _ _ ty => ty
  | Ecast _ ty => ty
  | Efield _ _ ty => ty
  end.

(** ** Statements *)

(** Clight statements are similar to those of Compcert C, with the addition
  of assigment (of a rvalue to a lvalue), assignment to a temporary,
  and function call (with assignment of the result to a temporary).
  The three C loops are replaced by a single infinite loop [Sloop s1
  s2] that executes [s1] then [s2] repeatedly.  A [continue] in [s1]
  branches to [s2]. *)

Definition label := ident.

Inductive statement : Type :=
  | Sskip : statement                   (**r do nothing *)
  | Sassign : expr -> expr -> statement (**r assignment [lvalue = rvalue] *)
  | Sset : ident -> expr -> statement   (**r assignment [tempvar = rvalue] *)
  | Scall: option ident -> expr -> list expr -> statement (**r function call *)
  | Sbuiltin: option ident -> external_function -> typelist -> list expr -> statement (**r builtin invocation *)
  | Ssequence : statement -> statement -> statement  (**r sequence *)
  | Sifthenelse : expr  -> statement -> statement -> statement (**r conditional *)
  | Sloop: statement -> statement -> statement (**r infinite loop *)
  | Sbreak : statement                      (**r [break] statement *)
  | Scontinue : statement                   (**r [continue] statement *)
  | Sreturn : option expr -> statement      (**r [return] statement *)
  | Sswitch : expr -> labeled_statements -> statement  (**r [switch] statement *)
  | Slabel : label -> statement -> statement
  | Sgoto : label -> statement

with labeled_statements : Type :=            (**r cases of a [switch] *)
  | LSdefault: statement -> labeled_statements
  | LScase: int -> statement -> labeled_statements -> labeled_statements.

(** The C loops are derived forms. *)

Definition Swhile (e: expr) (s: statement) :=
  Sloop (Ssequence (Sifthenelse e Sskip Sbreak) s) Sskip.

Definition Sdowhile (s: statement) (e: expr) :=
  Sloop s (Sifthenelse e Sskip Sbreak).

Definition Sfor (s1: statement) (e2: expr) (s3: statement) (s4: statement) :=
  Ssequence s1 (Sloop (Ssequence (Sifthenelse e2 Sskip Sbreak) s3) s4).

(** ** Functions *)

(** A function definition is composed of its return type ([fn_return]),
  the names and types of its parameters ([fn_params]), the names
  and types of its local variables ([fn_vars]), and the body of the
  function (a statement, [fn_body]). *)

Record function : Type := mkfunction {
  fn_id: ident;
  fn_return: type;
  fn_params: list (ident * type);
  fn_vars: list (ident * type);
  fn_temps: list (ident * type);
  fn_body: statement
}.

Definition var_names (vars: list(ident * type)) : list ident :=
  List.map (@fst ident type) vars.

(** Functions can either be defined ([Internal]) or declared as
  external functions ([External]). *)

Inductive fundef : Type :=
  | Internal: function -> fundef
  | External: external_function -> typelist -> type -> fundef.

(** The type of a function definition. *)

Definition type_of_function (f: function) : type :=
  Tfunction (type_of_params (fn_params f)) (fn_return f).

Definition type_of_fundef (f: fundef) : type :=
  match f with
  | Internal fd => type_of_function fd
  | External id args res => Tfunction args res
  end.

(** ** Programs *)

(** A program is a collection of named functions, plus a collection
  of named global variables, carrying their types and optional initialization 
  data.  See module [AST] for more details. *)

Definition program : Type := AST.program fundef type.

(** * Operational semantics *)

(** The semantics uses two environments.  The global environment
  maps names of functions and global variables to memory block references,
  and function pointers to their definitions.  (See module [Globalenvs].) *)

Definition genv := Genv.t fundef type.

(** The local environment maps local variables to block references and
  types.  The current value of the variable is stored in the
  associated memory block. *)

Definition env := PTree.t (block * type). (* map variable -> location & type *)

Definition empty_env: env := (PTree.empty (block * type)).

(** The temporary environment maps local temporaries to values. *)

Definition temp_env := PTree.t val.

(** [deref_loc ty m b ofs v] computes the value of a datum
  of type [ty] residing in memory [m] at block [b], offset [ofs].
  If the type [ty] indicates an access by value, the corresponding
  memory load is performed.  If the type [ty] indicates an access by
  reference or by copy, the pointer [Vptr b ofs] is returned. *)

Inductive deref_loc (ty: type) (m: mem) (b: block) (ofs: int) : val -> Prop :=
  | deref_loc_value: forall chunk v,
      access_mode ty = By_value chunk ->
      Mem.loadv chunk m (Vptr b ofs) = Some v ->
      forall NVUNDEF: v <> Vundef,
      forall ALIGN: (align_chunk chunk | Mem.option_tag_align (Mem.tget m b)),
      deref_loc ty m b ofs v
  | deref_loc_reference:
      access_mode ty = By_reference ->
      deref_loc ty m b ofs (Vptr b ofs)
  | deref_loc_copy:
      access_mode ty = By_copy ->
      deref_loc ty m b ofs (Vptr b ofs).

Lemma deref_loc_not_vundef:
  forall ty m b ofs v,
    deref_loc ty m b ofs v ->
    v <> Vundef.
Proof.
  inversion 1; congruence.
Qed.

(** Symmetrically, [assign_loc ty m b ofs v m'] returns the
  memory state after storing the value [v] in the datum
  of type [ty] residing in memory [m] at block [b], offset [ofs].
  This is allowed only if [ty] indicates an access by value or by copy.
  [m'] is the updated memory state. *)

Inductive assign_loc (ty: type) (m: mem) (b: block) (ofs: int):
                                            val -> mem -> Prop :=
  | assign_loc_value: forall v chunk m',
      access_mode ty = By_value chunk ->
      Mem.storev chunk m (Vptr b ofs) v = Some m' ->
      forall NVUNDEF: v <> Vundef,
      forall ALIGN: (align_chunk chunk | Mem.option_tag_align (Mem.tget m b)),
      assign_loc ty m b ofs v m'
  | assign_loc_copy: forall b' ofs' bytes m',
      access_mode ty = By_copy ->
      (alignof ty | Int.unsigned ofs') -> (alignof ty | Int.unsigned ofs) ->
      forall ALIGN: (alignof ty | Mem.option_tag_align (Mem.tget m b)),
      forall ALIGN': (alignof ty | Mem.option_tag_align (Mem.tget m b')),
      b' <> b \/ Int.unsigned ofs' = Int.unsigned ofs
              \/ Int.unsigned ofs' + sizeof ty <= Int.unsigned ofs
              \/ Int.unsigned ofs + sizeof ty <= Int.unsigned ofs' ->
      Mem.loadbytes m b' (Int.unsigned ofs') (sizeof ty) = Some bytes ->
      Mem.storebytes m b (Int.unsigned ofs) bytes = Some m' ->
      assign_loc ty m b ofs (Vptr b' ofs') m'.

Lemma assign_loc_not_vundef:
  forall ty m b ofs v m',
    assign_loc ty m b ofs v m' ->
    v <> Vundef.
Proof.
  inversion 1; congruence.
Qed.

(** Allocation of function-local variables.
  [alloc_variables e1 m1 vars e2 m2] allocates one memory block
  for each variable declared in [vars], and associates the variable
  name with this block.  [e1] and [m1] are the initial local environment
  and memory state.  [e2] and [m2] are the final local environment
  and memory state. *)

Inductive alloc_variables (fid: ident) (b: block): env -> mem ->
                           list (ident * type) ->
                           env -> mem -> Prop :=
  | alloc_variables_nil:
      forall e m,
      alloc_variables fid b e m nil e m
  | alloc_variables_cons:
      forall e m id ty vars m1 b1 m2 e2,
      Mem.alloc m 0 (sizeof ty) (Mem.Tag_stack (Mem.Stack_var (Mem.Svar fid id b (sizeof ty) (alignof ty) None))) = (m1, b1) ->
      alloc_variables fid b (PTree.set id (b1, ty) e) m1 vars e2 m2 ->
      alloc_variables fid b e m ((id, ty) :: vars) e2 m2.

(** Initialization of local variables that are parameters to a function.
  [bind_parameters e m1 params args m2] stores the values [args]
  in the memory blocks corresponding to the variables [params].
  [m1] is the initial memory state and [m2] the final memory state. *)

Inductive bind_parameters (e: env):
                           mem -> list (ident * type) -> list val ->
                           mem -> Prop :=
  | bind_parameters_nil:
      forall m,
      bind_parameters e m nil nil m
  | bind_parameters_cons:
      forall m id ty params v1 vl b m1 m2,
      PTree.get id e = Some(b, ty) ->
      assign_loc ty m b Int.zero v1 m1 ->
      bind_parameters e m1 params vl m2 ->
      bind_parameters e m ((id, ty) :: params) (v1 :: vl) m2.

(** Initialization of temporary variables *)

Fixpoint create_undef_temps (temps: list (ident * type)) : temp_env :=
  match temps with
  | nil => PTree.empty val
  | (id, t) :: temps' => PTree.set id Vundef (create_undef_temps temps')
 end.

(** Initialization of temporary variables that are parameters to a function. *)

Fixpoint bind_parameter_temps (formals: list (ident * type)) (args: list val)
                              (le: temp_env) : option temp_env :=
 match formals, args with
 | nil, nil => Some le
 | (id, t) :: xl, v :: vl => bind_parameter_temps xl vl (PTree.set id v le)
 | _, _ => None
 end. 

(** Return the list of blocks in the codomain of [e], with low and high bounds. *)
(*
Definition block_of_binding (id_b_ty: ident * (block * type)) :=
  match id_b_ty with (id, (b, ty)) => (b, 0, sizeof ty) end.

Definition blocks_of_env (e: env) : list (block * Z * Z) :=
  List.map block_of_binding (PTree.elements e).
*)

Function free_variables (b: block) (l: list (ident * type)) (m: mem) {struct l}: option mem :=
  match l with
    | nil => Some m
    | (_, ty) :: l' =>
      match Mem.free m b 0 (sizeof ty) with
        | Some m' => free_variables (Zsucc b) l' m'
        | _ => None
      end
  end.

(** Optional assignment to a temporary *)

Definition set_opttemp (optid: option ident) (v: val) (le: temp_env) :=
  match optid with
  | None => le
  | Some id => PTree.set id v le
  end.

(** Selection of the appropriate case of a [switch], given the value [n]
  of the selector expression. *)

Fixpoint select_switch (n: int) (sl: labeled_statements)
                       {struct sl}: labeled_statements :=
  match sl with
  | LSdefault _ => sl
  | LScase c s sl' => if Int.eq c n then sl else select_switch n sl'
  end.

(** Turn a labeled statement into a sequence *)

Fixpoint seq_of_labeled_statement (sl: labeled_statements) : statement :=
  match sl with
  | LSdefault s => s
  | LScase c s sl' => Ssequence s (seq_of_labeled_statement sl')
  end.

Section SEMANTICS.

Section SEM.
Variable ge: genv.

(** [type_of_global b] returns the type of the global variable or function
  at address [b]. *)

Definition type_of_global (b: block) : option type :=
  match Genv.find_var_info ge b with
  | Some gv => Some gv.(gvar_info)
  | None =>
      match Genv.find_funct_ptr ge b with
      | Some fd => Some(type_of_fundef fd)
      | None => None
      end
  end.

(** ** Evaluation of expressions *)

Section EXPR.

Variable e: env.
Variable le: temp_env.
Variable m: mem.

(** [eval_expr ge e m a v] defines the evaluation of expression [a]
  in r-value position.  [v] is the value of the expression.
  [e] is the current environment and [m] is the current memory state. *)

Inductive eval_expr: expr -> val -> Prop :=
  | eval_Econst_int:   forall i ty,
      eval_expr (Econst_int i ty) (Vint i)
  | eval_Econst_float:   forall f ty,
      eval_expr (Econst_float f ty) (Vfloat f)
  | eval_Etempvar:  forall id ty v,
      le!id = Some v ->
(*      forall NVUNDEF: ty <> Tvoid -> v <> Vundef, *)
      eval_expr (Etempvar id ty) v
  | eval_Eaddrof: forall a ty loc ofs,
      eval_lvalue a loc ofs ->
      eval_expr (Eaddrof a ty) (Vptr loc ofs)
  | eval_Eunop:  forall op a ty v1 v,
      eval_expr a v1 ->
      sem_unary_operation op v1 (typeof a) = Some v ->
      eval_expr (Eunop op a ty) v
  | eval_Ebinop: forall op a1 a2 ty v1 v2 v,
      eval_expr a1 v1 ->
      eval_expr a2 v2 ->
      sem_binary_operation op v1 (typeof a1) v2 (typeof a2) m = Some v ->
      eval_expr (Ebinop op a1 a2 ty) v
  | eval_Ecast:   forall a ty v1 v,
      eval_expr a v1 ->
      sem_cast v1 (typeof a) ty = Some v ->
(*      forall NVUNDEF: ty <> Tvoid -> v <> Vundef, *)
      eval_expr (Ecast a ty) v
  | eval_Elvalue: forall a loc ofs v,
      eval_lvalue a loc ofs -> 
      deref_loc (typeof a) m loc ofs v ->
      eval_expr a v

(** [eval_lvalue ge e m a b ofs] defines the evaluation of expression [a]
  in l-value position.  The result is the memory location [b, ofs]
  that contains the value of the expression [a]. *)

with eval_lvalue: expr -> block -> int -> Prop :=
  | eval_Evar_local:   forall id l ty,
      e!id = Some(l, ty) ->
      eval_lvalue (Evar id ty) l Int.zero
  | eval_Evar_global: forall id l ty,
      e!id = None ->
      Genv.find_symbol ge id = Some l ->
      type_of_global l = Some ty ->
      eval_lvalue (Evar id ty) l Int.zero
  | eval_Ederef: forall a ty l ofs,
      eval_expr a (Vptr l ofs) ->
      eval_lvalue (Ederef a ty) l ofs
 | eval_Efield_struct:   forall a i ty l ofs id fList att delta,
      eval_expr a (Vptr l ofs) ->
      typeof a = Tstruct id fList att ->
      field_offset i fList = OK delta ->
      eval_lvalue (Efield a i ty) l (Int.add ofs (Int.repr delta))
 | eval_Efield_union:   forall a i ty l ofs id fList att,
      eval_expr a (Vptr l ofs) ->
      typeof a = Tunion id fList att ->
      eval_lvalue (Efield a i ty) l ofs.

Scheme eval_expr_ind2 := Minimality for eval_expr Sort Prop
  with eval_lvalue_ind2 := Minimality for eval_lvalue Sort Prop.
Combined Scheme eval_expr_lvalue_ind from eval_expr_ind2, eval_lvalue_ind2.

(** [eval_exprlist ge e m al tyl vl] evaluates a list of r-value
  expressions [al], cast their values to the types given in [tyl],
  and produces the list of cast values [vl].  It is used to
  evaluate the arguments of function calls. *)

Inductive eval_exprlist: list expr -> typelist -> list val -> Prop :=
  | eval_Enil:
      eval_exprlist nil Tnil nil
  | eval_Econs:   forall a bl ty tyl v1 v2 vl,
      eval_expr a v1 ->
      sem_cast v1 (typeof a) ty = Some v2 ->
      eval_exprlist bl tyl vl ->
      forall NVUNDEF: (* ty <> Tvoid -> *) v2 <> Vundef,
      eval_exprlist (a :: bl) (Tcons ty tyl) (v2 :: vl).

End EXPR.

(** ** Transition semantics for statements and functions *)

(** Continuations *)

Inductive cont: Type :=
  | Kstop: cont
  | Kseq: statement -> cont -> cont       (**r [Kseq s2 k] = after [s1] in [s1;s2] *)
  | Kloop1: statement -> statement -> cont -> cont (**r [Kloop1 s1 s2 k] = after [s1] in [Sloop s1 s2] *)
  | Kloop2: statement -> statement -> cont -> cont (**r [Kloop1 s1 s2 k] = after [s2] in [Sloop s1 s2] *)
  | Kswitch: cont -> cont       (**r catches [break] statements arising out of [switch] *)
  | Kcall: block ->
           option ident ->                  (**r where to store result *)
           function ->                      (**r calling function *)
           env ->                           (**r local env of calling function *)
           temp_env ->                      (**r temporary env of calling function *)
           cont -> cont.

(** Pop continuation until a call or stop *)

Fixpoint call_cont (k: cont) : cont :=
  match k with
  | Kseq s k => call_cont k
  | Kloop1 s1 s2 k => call_cont k
  | Kloop2 s1 s2 k => call_cont k
  | Kswitch k => call_cont k
  | _ => k
  end.

Definition is_call_cont (k: cont) : Prop :=
  match k with
  | Kstop => True
  | Kcall _ _ _ _ _ _ => True
  | _ => False
  end.

Function append_cont (k1 k2: cont) {struct k1}: cont :=
  match k1 with
    | Kstop => k2
    | Kseq s k => Kseq s (append_cont k k2)
    | Kloop1 s1 s2 k => Kloop1 s1 s2 (append_cont k k2)
    | Kloop2 s1 s2 k => Kloop2 s1 s2 (append_cont k k2)
    | Kswitch k => Kswitch (append_cont k k2)
    | Kcall b oi fn e te k => Kcall b oi fn e te (append_cont k k2)
  end.

Function cont_of_stack (l: list cont): cont :=
  match l with
    | nil => Kstop
    | k :: l => append_cont k (cont_of_stack l)
  end.

Function no_calls (k: cont): bool :=
  match k with
    | Kcall _ _ _ _ _ _ => false
    | Kstop => true
    | Kseq _ k => no_calls k
    | Kloop1 _ _ k => no_calls k
    | Kloop2 _ _ k => no_calls k
    | Kswitch k => no_calls k
  end.

Function header_cont (k: cont): cont :=
  match k with
    | Kstop => Kstop
    | Kcall _ _ _ _ _ _ => Kstop
    | Kseq s k => Kseq s (header_cont k)
    | Kloop1 s1 s2 k => (Kloop1 s1 s2 (header_cont k))
    | Kloop2 s1 s2 k => (Kloop2 s1 s2 (header_cont k))
    | Kswitch k => (Kswitch (header_cont k))
  end.

Lemma header_cont_no_calls: forall k,
                              no_calls (header_cont k) = true.
Proof.
  induction k; simpl; auto.
Qed.

Lemma header_call_cont_append: forall k, append_cont (header_cont k) (call_cont k) = k.
Proof.
  induction k; simpl; congruence.
Qed.

Function cont_stackframe (k: cont): bool :=
  match k with
    | Kcall _ _ _ _ _ k => no_calls k
    | _ => false
  end.

Lemma cont_decomp:
  forall k,
  {hl : _ * _ | match hl with (h, l) =>
              k = append_cont h (cont_of_stack l) /\
              no_calls h = true /\
              forall k', In k' l -> cont_stackframe k' = true
                end}.
Proof.
  induction k; simpl.
  (* Kstop *)
   exists (Kstop, nil). simpl. tauto.
  (* Kseq *)
   destruct IHk.
   destruct x.
   destruct y.
   exists (Kseq s c, l).
   split.
    simpl. congruence.
   simpl. tauto.
   (* Kloop1 *)
   destruct IHk. destruct x. destruct y.
   exists (Kloop1 s s0 c, l).
   simpl. split; auto. congruence.
   (* Kloop2 *)
   destruct IHk. destruct x. destruct y.
   exists (Kloop2 s s0 c, l).
   simpl. split; auto. congruence.
   (* Kswitch *)
   destruct IHk. destruct x. destruct y.
   exists (Kswitch c, l).
   simpl. split; auto. congruence.
   (* Kcall *)
   destruct IHk. destruct x. destruct y. destruct H0.
   exists (Kstop, Kcall b o f e t c :: l).
   simpl. split. congruence.
   split; auto.
   destruct 1; auto.
   subst. assumption.
Defined.

Lemma stack_is_call_cont:
  forall l,
    (forall k', In k' l -> cont_stackframe k' = true) ->
    is_call_cont (cont_of_stack l).
Proof.
  destruct l; simpl.
   tauto.
  intros. exploit H. left. reflexivity.
  functional inversion 1; simpl. tauto.
Qed.

Lemma is_call_cont_call_cont:
  forall k1,
    no_calls k1 = true ->
    forall k2,
      is_call_cont k2 ->
      call_cont (append_cont k1 k2) = k2.
Proof.
  induction k1; simpl; try assumption.
   intros. destruct k2; try contradiction; reflexivity.
  discriminate.
Qed.

(** States *)

Inductive state `{M: Mem.MEM mem}: Type :=
  | State
      (bf: block)
      (f: function)
      (s: statement)
      (k: cont)
      (e: env)
      (le: temp_env)
      (m: mem) : state
  | Callstate
      (fd: fundef)
      (args: list val)
      (k: cont)
      (m: mem) : state
  | Returnstate
      (res: val)
      (k: cont)
      (m: mem) : state.
                 
(** Find the statement and manufacture the continuation 
  corresponding to a label *)

Fixpoint find_label (lbl: label) (s: statement) (k: cont) 
                    {struct s}: option (statement * cont) :=
  match s with
  | Ssequence s1 s2 =>
      match find_label lbl s1 (Kseq s2 k) with
      | Some sk => Some sk
      | None => find_label lbl s2 k
      end
  | Sifthenelse a s1 s2 =>
      match find_label lbl s1 k with
      | Some sk => Some sk
      | None => find_label lbl s2 k
      end
  | Sloop s1 s2 =>
      match find_label lbl s1 (Kloop1 s1 s2 k) with
      | Some sk => Some sk
      | None => find_label lbl s2 (Kloop2 s1 s2 k)
      end
  | Sswitch e sl =>
      find_label_ls lbl sl (Kswitch k)
  | Slabel lbl' s' =>
      if ident_eq lbl lbl' then Some(s', k) else find_label lbl s' k
  | _ => None
  end

with find_label_ls (lbl: label) (sl: labeled_statements) (k: cont) 
                    {struct sl}: option (statement * cont) :=
  match sl with
  | LSdefault s => find_label lbl s k
  | LScase _ s sl' =>
      match find_label lbl s (Kseq (seq_of_labeled_statement sl') k) with
      | Some sk => Some sk
      | None => find_label_ls lbl sl' k
      end
  end.

(** Semantics for allocation of variables and binding of parameters at
  function entry.  Two semantics are supported: one where
  parameters are local variables, reside in memory, and can have their address
  taken; the other where parameters are temporary variables and do not reside
  in memory.  We parameterize the [step] transition relation over the
  parameter binding semantics, then instantiate it later to give the two
  semantics described above. *)

Variable function_entry: function -> list val -> mem -> env -> temp_env -> mem -> block -> Prop.

(** Transition relation *)

Inductive step: state -> trace -> state -> Prop :=

  | step_assign:   forall bf f a1 a2 k e le m loc ofs v2 v m',
      eval_lvalue e le m a1 loc ofs ->
      eval_expr e le m a2 v2 ->
      sem_cast v2 (typeof a2) (typeof a1) = Some v ->
      assign_loc (typeof a1) m loc ofs v m' ->
      step (State bf f (Sassign a1 a2) k e le m)
        E0 (State bf f Sskip k e le m')

  | step_set:   forall bf f id a k e le m v,
      eval_expr e le m a v ->
(*      forall NVUNDEF: v <> Vundef, *)
      step (State bf f (Sset id a) k e le m)
        E0 (State bf f Sskip k e (PTree.set id v le) m)

  | step_call:   forall bf f optid a al k e le m tyargs tyres vf vargs fd,
      classify_fun (typeof a) = fun_case_f tyargs tyres ->
      eval_expr e le m a vf ->
      eval_exprlist e le m al tyargs vargs ->
      Genv.find_funct ge vf = Some fd ->
      type_of_fundef fd = Tfunction tyargs tyres ->
      step (State bf f (Scall optid a al) k e le m)
        E0 (Callstate fd vargs (Kcall bf optid f e le k) m)

  | step_builtin:   forall bf f optid ef tyargs al k e le m vargs t vres m',
      eval_exprlist e le m al tyargs vargs ->
      external_call ef ge vargs m t vres m' ->
(*      forall NVUNDEF: optid <> None -> vres <> Vundef, *)
      step (State bf f (Sbuiltin optid ef tyargs al) k e le m)
         t (State bf f Sskip k e (set_opttemp optid vres le) m')

  | step_seq:  forall bf f s1 s2 k e le m,
      step (State bf f (Ssequence s1 s2) k e le m)
        E0 (State bf f s1 (Kseq s2 k) e le m)
  | step_skip_seq: forall bf f s k e le m,
      step (State bf f Sskip (Kseq s k) e le m)
        E0 (State bf f s k e le m)
  | step_continue_seq: forall bf f s k e le m,
      step (State bf f Scontinue (Kseq s k) e le m)
        E0 (State bf f Scontinue k e le m)
  | step_break_seq: forall bf f s k e le m,
      step (State bf f Sbreak (Kseq s k) e le m)
        E0 (State bf f Sbreak k e le m)

  | step_ifthenelse:  forall bf f a s1 s2 k e le m v1 b,
      eval_expr e le m a v1 ->
      bool_val v1 (typeof a) = Some b ->
      step (State bf f (Sifthenelse a s1 s2) k e le m)
        E0 (State bf f (if b then s1 else s2) k e le m)

  | step_loop: forall bf f s1 s2 k e le m,
      step (State bf f (Sloop s1 s2) k e le m)
        E0 (State bf f s1 (Kloop1 s1 s2 k) e le m)
  | step_skip_or_continue_loop1:  forall bf f s1 s2 k e le m x,
      x = Sskip \/ x = Scontinue ->
      step (State bf f x (Kloop1 s1 s2 k) e le m)
        E0 (State bf f s2 (Kloop2 s1 s2 k) e le m)
  | step_break_loop1:  forall bf f s1 s2 k e le m,
      step (State bf f Sbreak (Kloop1 s1 s2 k) e le m)
        E0 (State bf f Sskip k e le m)
  | step_skip_loop2: forall bf f s1 s2 k e le m,
      step (State bf f Sskip (Kloop2 s1 s2 k) e le m)
        E0 (State bf f (Sloop s1 s2) k e le m)
  | step_break_loop2: forall bf f s1 s2 k e le m,
      step (State bf f Sbreak (Kloop2 s1 s2 k) e le m)
        E0 (State bf f Sskip k e le m)

  | step_return_0: forall bf f k e le m m',      
      free_variables (Zsucc bf) f.(fn_vars) m = Some m' ->
      step (State bf f (Sreturn None) k e le m)
        E0 (Returnstate Vundef (call_cont k) m')
  | step_return_1: forall bf f a k e le m v v' m',
      eval_expr e le m a v -> 
      sem_cast v (typeof a) f.(fn_return) = Some v' ->
      free_variables (Zsucc bf) f.(fn_vars) m = Some m' ->
(*      forall NVUNDEF: f.(fn_return) <> Tvoid -> v' <> Vundef, *)
      step (State bf f (Sreturn (Some a)) k e le m)
        E0 (Returnstate v' (call_cont k) m')
  | step_skip_call: forall bf f k e le m m',
      is_call_cont k ->
      free_variables (Zsucc bf) f.(fn_vars) m = Some m' ->
      step (State bf f Sskip k e le m)
        E0 (Returnstate Vundef k m')

  | step_switch: forall bf f a sl k e le m n,
      eval_expr e le m a (Vint n) ->
      step (State bf f (Sswitch a sl) k e le m)
        E0 (State bf f (seq_of_labeled_statement (select_switch n sl)) (Kswitch k) e le m)
  | step_skip_break_switch: forall bf f x k e le m,
      x = Sskip \/ x = Sbreak ->
      step (State bf f x (Kswitch k) e le m)
        E0 (State bf f Sskip k e le m)
  | step_continue_switch: forall bf f k e le m,
      step (State bf f Scontinue (Kswitch k) e le m)
        E0 (State bf f Scontinue k e le m)

  | step_label: forall bf f lbl s k e le m,
      step (State bf f (Slabel lbl s) k e le m)
        E0 (State bf f s k e le m)

  | step_goto: forall bf f lbl k e le m s' k',
      find_label lbl f.(fn_body) (call_cont k) = Some (s', k') ->
      step (State bf f (Sgoto lbl) k e le m)
        E0 (State bf f s' k' e le m)

  | step_internal_function: forall bf f vargs k m e le m1,
      function_entry f vargs m e le m1 bf ->
      step (Callstate (Internal f) vargs k m)
        E0 (State bf f f.(fn_body) k e le m1)

  | step_external_function: forall ef targs tres vargs k m vres t m',
      external_call ef ge vargs m t vres m' ->
      step (Callstate (External ef targs tres) vargs k m)
         t (Returnstate vres k m')

  | step_returnstate: forall bf v optid f e le k m,
(*                        forall NVUNDEF: optid <> None -> v <> Vundef, *)
      step (Returnstate v (Kcall bf optid f e le k) m)
        E0 (State bf f Sskip k e (set_opttemp optid v le) m).

(** ** Whole-program semantics *)

(** Execution of whole programs are described as sequences of transitions
  from an initial state to a final state.  An initial state is a [Callstate]
  corresponding to the invocation of the ``main'' function of the program
  without arguments and with an empty continuation. *)

Inductive initial_state (p: program): state -> Prop :=
  | initial_state_intro: forall b f m0,
      let ge := Genv.globalenv p in
      Genv.init_mem p = Some m0 ->
      Genv.find_symbol ge p.(prog_main) = Some b ->
      Genv.find_funct_ptr ge b = Some f ->
      type_of_fundef f = Tfunction Tnil type_int32s ->
      initial_state p (Callstate f nil Kstop m0).

(** A final state is a [Returnstate] with an empty continuation. *)

Inductive final_state: state -> int -> Prop :=
  | final_state_intro: forall r m,
      final_state (Returnstate (Vint r) Kstop m) r.

End SEM.

Definition semantics function_entry (p: program) :=
  Semantics (fun ge => step ge (function_entry ge)) (initial_state p) final_state (Genv.globalenv p).

Lemma semantics_receptive_general:
  forall function_entry (p: program), receptive (semantics function_entry p).
Proof.
  intros. constructor; simpl; intros.
(* receptiveness *)
  assert (t1 = E0 -> exists s2, step (Genv.globalenv p) (function_entry (Genv.globalenv p)) s t2 s2).
    intros. subst. inv H0. exists s1; auto.
  inversion H; subst; auto.
  (* builtin *)
  exploit external_call_receptive; eauto. intros [vres2 [m2 EC2]]. 
  econstructor; econstructor; eauto.
  (* external *)
  exploit external_call_receptive; eauto. intros [vres2 [m2 EC2]]. 
  exists (Returnstate vres2 k m2). econstructor; eauto.
(* trace length *)
  red; intros. inv H; simpl; try omega.
  eapply external_call_trace_length; eauto.
  eapply external_call_trace_length; eauto.
Qed.

End SEMANTICS.


(** The two semantics for function parameters.  First, parameters as local variables. *)

Inductive function_entry1 (ge: genv) (f: function) (vargs: list val) (m: mem) (e: env) (le: temp_env) (m': mem) (b: block) : Prop :=
  | function_entry1_intro: 
      list_norepet (var_names f.(fn_params) ++ var_names f.(fn_vars)) ->
      forall m0
             (HDUMMY: Mem.alloc m 0 0 (Mem.Tag_stack (Mem.Stack_main f.(fn_id) None)) = (m0, b))
             m1,
      alloc_variables f.(fn_id) b empty_env m0 (f.(fn_params) ++ f.(fn_vars)) e m1 ->
      bind_parameters e m1 f.(fn_params) vargs m' ->
      le = create_undef_temps f.(fn_temps) ->
      function_entry1 ge f vargs m e le m' b.

Definition step1 (ge: genv) := step ge (function_entry1 ge).

(** Second, parameters as temporaries. *)

Inductive function_entry2 (ge: genv) (f: function) (vargs: list val) (m: mem) (e: env) (le: temp_env) (m': mem) (b: block) : Prop :=
  | function_entry2_intro:
      forall bf (Hsymb: Genv.find_symbol ge f.(fn_id) = Some bf) (Hblock: Genv.find_funct_ptr ge bf = Some (Internal f)),
      list_norepet (var_names f.(fn_vars)) ->
      list_norepet (var_names f.(fn_params)) ->
      list_disjoint (var_names f.(fn_params)) (var_names f.(fn_temps)) ->
      forall m0 
             (HDUMMY: Mem.alloc m 0 0 (Mem.Tag_stack (Mem.Stack_main f.(fn_id) None)) = (m0, b)),
      alloc_variables f.(fn_id) b empty_env m0 f.(fn_vars) e m' ->
      bind_parameter_temps f.(fn_params) vargs (create_undef_temps f.(fn_temps)) = Some le ->
      function_entry2 ge f vargs m e le m' b.

Definition step2 (ge: genv) := step ge (function_entry2 ge).

(** Wrapping up these definitions in two small-step semantics. *)

Definition semantics1 := semantics function_entry1.

Definition semantics2 := semantics function_entry2.

(** This semantics is receptive to changes in events. *)

Lemma semantics_receptive:
  forall (p: program), receptive (semantics1 p).
Proof.
  eauto using semantics_receptive_general.
Qed.

Lemma semantics2_receptive:
  forall (p: program), receptive (semantics2 p).
Proof.
  eauto using semantics_receptive_general.
Qed.

End WITHMEM.
