Require Import Coqlib.
Require Import Values.
Require Import Integers.
Require Import Maps.
Require Import AST.
Require Import Memory.
Require Import StrongInject.
Require Memimpl.

Require Import Coq.Classes.Morphisms.
Require Import Coq.Relations.Relations.

Existing Instance Memimpl.mem_MEM.
Open Scope Z_scope.

(** MathClasses-style "Decision" class. Used for [assert] below. *)

Class Decision (P: Prop) := { decide: {P} + {~P} }.
Arguments decide P {_}.

Instance decide_Zeq (m n: Z): Decision (m = n) := {decide := Z_eq_dec m n}.
Instance decide_Zle (m n: Z): Decision (m <= n) := {decide := Z_le_dec m n}.
Instance decide_Zlt (m n: Z): Decision (m < n) := {decide := Z_lt_dec m n}.
Instance decide_Zge (m n: Z): Decision (m >= n) := {decide := Z_ge_dec m n}.
Instance decide_Zgt (m n: Z): Decision (m > n) := {decide := Z_gt_dec m n}.
Instance decide_nateq m n: Decision (m = n) := { decide := eq_nat_dec m n }.

Instance: forall A B, Decision A -> Decision B -> Decision (A /\ B) := {
  decide :=
    match (decide A) with
      | left HA =>
          match (decide B) with
            | left HB => left (conj HA HB)
            | right HnB => right (fun H => HnB (proj2 H))
          end
      | right HnA => right (fun H => HnA (proj1 H))
    end
}.

Program Instance decide_not_none {A} (a: option A): Decision (a <> None) := {
  decide :=
    match a with
      | Some _ => left _
      | None => right _
    end
}.

Section DECIDE_PROD.
  Context A `{Adec: forall x y: A, Decision (x = y)}.
  Context B `{Bdec: forall x y: B, Decision (x = y)}.

  Global Instance decide_eq_pair: forall (x y: A * B), Decision (x = y).
  Proof.
    intros [x1 x2] [y1 y2].
    constructor.
    destruct (decide (x1 = y1)).
    destruct (decide (x2 = y2)).
    left; congruence.
    right; intro H; inversion H; now auto.
    right; intro H; inversion H; now auto.
  Defined.
End DECIDE_PROD.

(** * Equality of memory states *)

Section MEMEQ.
  Require Import Axioms.
  Import Memimpl.

  Record block_eq b (m1 m2: mem) := {
    block_eq_contents:
      ZMap'.get b (mem_contents m1) =
      ZMap'.get b (mem_contents m2);
    block_eq_access:
      ZMap'.get b (mem_access m1) =
      ZMap'.get b (mem_access m2);
    block_eq_tag:
      ZMap'.get b (tget m1) =
      ZMap'.get b (tget m2);
    block_eq_bounds:
      ZMap'.get b (block_bounds m1) =
      ZMap'.get b (block_bounds m2)
  }.

  Global Instance block_eq_equiv b: Equivalence (block_eq b).
  Proof.
    split.
    - intro m.
      split; reflexivity.
    - intros m1 m2 Hm.
      destruct Hm.
      split; symmetry; assumption.
    - intros m1 m2 m3 H12 H23.
      destruct H12, H23.
      split; etransitivity; eassumption.
  Qed.

  Lemma block_eq_alloc b m lo hi t m' b':
    b < nextblock m ->
    Mem.alloc m lo hi t = (m', b') ->
    block_eq b m m'.
  Proof.
    intros Hb Hm'.
    assert (Hb': b' = Mem.nextblock m) by (eapply Mem.alloc_result; eassumption).
    simpl in *; unfold talloc in Hm'.
    destruct (talloc_strong m lo hi t) as [m1 [b1 [m2 H]]].
    destruct H as [H1 [H2 [H3 [H4 [H5 H6]]]]].
    inversion Hm'; subst; clear Hm'.
    transitivity m2.
    * clear m' H2 H3 H4 H5 H6.
      unfold alloc in H1.
      inversion H1; clear H0 H1.
      split;
      simpl in *;
      try rewrite ZMap'.gsspec;
      unfold ZMap'.get, ZMap'.elt_eq, ZEQ.eq;
      destruct (zeq b (nextblock m)) as [Hb'|Hb'];
      try omega;
      try reflexivity.
    * split; try congruence.
      unfold ZMap'.get.
      rewrite H5.
      destruct (zeq b (nextblock m)).
      - omega.
      - reflexivity.
  Qed.

  Definition mem_eq thr (m1 m2: mem) :=
    forall b, b < thr -> block_eq b m1 m2.

  Global Instance mem_eq_equiv thr: Equivalence (mem_eq thr).
  Proof.
    split.
    - intros m b Hb.
      reflexivity.
    - intros m1 m2 Hm b Hb.
      symmetry.
      now auto.
    - intros m1 m2 m3 H12 H23 b Hb.
      transitivity m2;
      now auto.
  Qed.

  Lemma mem_eq_succ thr m1 m2:
    block_eq thr m1 m2 ->
    mem_eq thr m1 m2 ->
    mem_eq (thr + 1) m1 m2.
  Proof.
    intros Hthr H b Hb.
    destruct (decide (b = thr)) as [Hbt|Hbt].
    - subst; assumption.
    - apply H; omega.
  Qed.

  Lemma mem_eq_alloc thr m lo hi t m' b':
    thr <= nextblock m ->
    Mem.alloc m lo hi t = (m', b') ->
    mem_eq thr m m'.
  Proof.
    intros Hthr Hm' b Hb.
    eapply block_eq_alloc.
    - omega.
    - eassumption.
  Qed.

  Lemma mem_ext_eq (m1 m2: mem):
    mem_contents m1 = mem_contents m2 ->
    mem_access m1 = mem_access m2 ->
    nextblock m1 = nextblock m2 ->
    tget m1 = tget m2 ->
    block_bounds m1 = block_bounds m2 ->
    m1 = m2.
  Proof.
    destruct m1 as [c1 a1 nb1 Hnb1 Ha1 Hcnb1 Hanb1 t1 Htnb1 b1 Hab1].
    destruct m2 as [c2 a2 nb2 Hnb2 Ha2 Hcnb2 Hanb2 t2 Htnb2 b2 Hab2].
    simpl.
    intros Hc Ha Hnb Ht Hb.
    subst.
    f_equal; apply proof_irr.
  Qed.

  Lemma mem_eq_eq n m1 m2:
    nextblock m1 = n ->
    nextblock m2 = n ->
    mem_eq n m1 m2 ->
    m1 = m2.
  Proof.
    intros H1 H2 H.
    apply mem_ext_eq.
    * (* contents *)
      apply functional_extensionality; intro b.
      destruct (decide (b < n)) as [Hb|Hb].
      - apply block_eq_contents, H, Hb.
      - apply functional_extensionality; intro ofs.
        transitivity Undef.
        apply nextblock_nocontents. rewrite H1. omega.
        symmetry.
        apply nextblock_nocontents. rewrite H2. omega.
    * (* access *)
      apply functional_extensionality; intro b.
      destruct (decide (b < n)) as [Hb|Hb].
      - apply block_eq_access, H, Hb.
      - apply functional_extensionality; intro ofs.
        apply functional_extensionality; intro k.
        transitivity (@None permission).
        apply nextblock_noaccess. rewrite H1. omega.
        symmetry.
        apply nextblock_noaccess. rewrite H2. omega.
    * (* nextblock *)
      congruence.
    * (* tget *)
      apply functional_extensionality; intro b.
      destruct (decide (b < n)) as [Hb|Hb].
      - apply block_eq_tag, H, Hb.
      - transitivity (@None Mem.tag).
        apply nextblock_notget. rewrite H1. omega.
        symmetry.
        apply nextblock_notget. rewrite H2. omega.
    * (* block_bounds *)
      apply functional_extensionality; intro b.
      destruct (decide (b < n)) as [Hb|Hb].
      - apply block_eq_bounds, H, Hb.
      - transitivity (0, 0).
        apply nextblock_nobounds. rewrite H1. omega.
        symmetry.
        apply nextblock_nobounds. rewrite H2. omega.
  Qed.
End MEMEQ.

(** Definitions and syntax for the "option" monad *)

Definition ret {A} (a: A): option A :=
  Some a.

Definition bind {A B} (f: A -> option B) (x: option A): option B :=
  match x with
    | Some a => (f a)
    | None => None
  end.

Notation "'do' x <- M ; N" := (bind (fun x => N) M)
  (at level 200, x ident, M at level 100, N at level 200).

Definition fmap {A B} (f: A -> B) (oa: option A): option B :=
  do a <- oa; ret (f a).

Definition assert P `{HP: Decision P}: option P :=
  match (decide P) with
    | left HP => Some HP
    | right _ => None
  end.

Lemma assert_inv `{Decision} {A} (f: P -> option A) (r: A):
  bind f (assert P) = Some r ->
  exists H, f H = Some r.
Proof.
  unfold assert.
  destruct (decide P); try discriminate.
  simpl.
  eauto.
Qed.

Ltac assert_inv H :=
  match type of H with
    | bind _ (assert _) = _ =>
        apply assert_inv in H;
        let H' := fresh H in
          destruct H as [H' H];
        try assert_inv H
  end.

Lemma monad_inv {A B} (f: A -> option B) (x: option A) (b: B):
  bind f x = ret b ->
  { a: A & x = ret a /\ f a = ret b }.
Proof.
  destruct x; try discriminate; now eauto.
Qed.

Ltac monad_inv H :=
  eapply monad_inv in H;
  let x := fresh "x" in
  let H' := fresh H in
  destruct H as [x [H' H]];
  try monad_inv H.

Ltac destruct_tget m b :=
  destruct (Mem.tget m b) as [[h | [f [sz|] | [fi vi r sz al [ofs|]]]] | ];
  try discriminate.

(** * Applying a per-block transformation to a memory state *)

Section MEMMAP.
  Import Memimpl.

  (** ** Individual memory blocks *)

  Record mem_block: Type := {
    mem_block_contents: ZMap'.t memval;
    mem_block_access (ofs: Z) (k: perm_kind): option permission;
    mem_block_tag: Mem.tag;
    mem_block_bounds: Z * Z
  }.

  Class mem_block_invariants (blk: mem_block): Prop := {
    mem_block_access_max ofs:
      perm_order'' (mem_block_access blk ofs Max) (mem_block_access blk ofs Cur);
    mem_block_bounds_perm lo hi ofs k:
      mem_block_bounds blk = (lo, hi) ->
      ~ (lo <= ofs < hi) ->
      mem_block_access blk ofs k = None
  }.

  Lemma mem_block_ext_eq blk1 blk2:
    mem_block_contents blk1 = mem_block_contents blk2 ->
    mem_block_access blk1 = mem_block_access blk2 ->
    mem_block_tag blk1 = mem_block_tag blk2 ->
    mem_block_bounds blk1 = mem_block_bounds blk2 ->
    blk1 = blk2.
  Proof.
    destruct blk1 as [c1 a1 t1 b1 Ha1 Hb1].
    destruct blk2 as [c2 a2 t2 b2 Ha2 Hb2].
    simpl.
    intros; subst.
    f_equal; now apply proof_irr.
  Qed.

  (** ** Block-based representation of memory. *)

  Record bbmem := {
    bbmem_nextblock: block;
    bbmem_block: block -> option mem_block
  }.

  Class bbmem_invariants (bbm: bbmem): Prop := {
    bbmem_nextblock_pos:
      bbmem_nextblock bbm > 0;
    bbmem_nextblock_noblock b:
      b >= bbmem_nextblock bbm <-> bbmem_block bbm b = None;
    bbmem_negblock_default b blk:
      b < 1 -> bbmem_block bbm b = Some blk ->
        mem_block_tag blk = (Mem.Tag_heap Mem.Heap_global);
    bbmem_block_invariants b blk :>
      bbmem_block bbm b = Some blk -> mem_block_invariants blk
  }.

  Lemma bbmem_ext_eq bbm1 bbm2
    `{Hbbm1: bbmem_invariants bbm1}
    `{Hbbm2: bbmem_invariants bbm2}:
    (forall b, bbmem_block bbm1 b = bbmem_block bbm2 b) ->
    bbm1 = bbm2.
  Proof.
    intro H.
    assert (Hnb: bbmem_nextblock bbm1 = bbmem_nextblock bbm2).
    * destruct (Ztrichotomy (bbmem_nextblock bbm1) (bbmem_nextblock bbm2))
                          as [Hlt | [Heq | Hgt]].
      - exfalso.
        specialize (H (bbmem_nextblock bbm1)).
        assert (H1:   bbmem_nextblock bbm1 >= bbmem_nextblock bbm1) by omega.
        assert (H2: ~ bbmem_nextblock bbm1 >= bbmem_nextblock bbm2) by omega.
        rewrite bbmem_nextblock_noblock in H1.
        rewrite bbmem_nextblock_noblock in H2.
        congruence.
      - assumption.
      - exfalso.
        specialize (H (bbmem_nextblock bbm2)).
        assert (H1: ~ bbmem_nextblock bbm2 >= bbmem_nextblock bbm1) by omega.
        assert (H2:   bbmem_nextblock bbm2 >= bbmem_nextblock bbm2) by omega.
        rewrite bbmem_nextblock_noblock in H1.
        rewrite bbmem_nextblock_noblock in H2.
        congruence.
    * assert (Hblk: bbmem_block bbm1 = bbmem_block bbm2)
        by (apply functional_extensionality; assumption).
      destruct bbm1, bbm2.
      simpl in *.
      subst.
      f_equal; apply proof_irr.
  Qed.

  (** ** Correspondence between [mem] and [bbmem] *)

  Definition get_block (m: mem) (b: block): option mem_block :=
    do t <- tget m b;
    ret {|
      mem_block_contents := mem_contents m b;
      mem_block_access := mem_access m b;
      mem_block_tag := t;
      mem_block_bounds := block_bounds m b
    |}.

  Global Instance get_block_invariants m b blk:
    get_block m b = Some blk -> mem_block_invariants blk.
  Proof.
    unfold get_block.
    destruct (tget m b) as [t|]; try discriminate; simpl.
    intro H; inversion H; subst; clear H.
    split; simpl.
    * intro ofs.
      apply access_max.
    * intros lo hi ofs k Hb Hofs.
      eapply block_bounds_perm.
      unfold ZMap'.get.
      destruct (block_bounds m b).
      eassumption.
      assumption.
  Qed.

  Lemma get_block_none m b:
    b >= nextblock m -> get_block m b = None.
  Proof.
    intro.
    unfold get_block.
    apply nextblock_notget in H.
    rewrite H.
    reflexivity.
  Qed.

  Lemma get_block_some m b:
    ~ b >= nextblock m -> exists blk, get_block m b = Some blk.
  Proof.
    intro.
    unfold get_block.
    rewrite nextblock_notget in H.
    destruct (tget m b).
    eexists.
    reflexivity.
    elim H.
    reflexivity.
  Qed.

  Definition bbmem_of_mem (m: mem): bbmem := {|
    bbmem_nextblock := nextblock m;
    bbmem_block b := get_block m b
  |}.

  Global Instance bbmem_of_mem_invariants (m:mem):
    bbmem_invariants (bbmem_of_mem m) | 10 := {}.
  Proof.
    + apply nextblock_pos.
    + simpl.
      unfold get_block.
      split; intro H.
      * assert (Ht: tget m b = None) by (apply nextblock_notget; assumption).
        rewrite Ht.
        reflexivity.
      * apply nextblock_notget.
        destruct (tget m b); try discriminate.
        reflexivity.
    + intros.
      inv H0.
      unfold get_block in H2.
      rewrite nextblock_default in H2; simpl in *.
      inv H2.
      reflexivity.
      assumption.
  Qed.

  Program Definition mem_of_bbmem (bbm: bbmem) `{Hbbm: bbmem_invariants bbm}: mem :=
  {|
    nextblock := bbmem_nextblock bbm;
    mem_contents b :=
      match bbmem_block bbm b with
        | Some blk => mem_block_contents blk
        | None => (fun _ => Undef)
      end;
    mem_access b :=
      match bbmem_block bbm b with
        | Some blk => mem_block_access blk
        | None => (fun _ _ => None)
      end;
    tget b :=
      do blk <- bbmem_block bbm b; ret (mem_block_tag blk);
    block_bounds b :=
      match bbmem_block bbm b with
        | Some blk => mem_block_bounds blk
        | None => (0, 0)
      end
  |}.
  Next Obligation.
    apply bbmem_nextblock_pos.
  Qed.
  Next Obligation.
    pose proof (bbmem_block_invariants b) as Hbi.
    unfold ZMap'.get.
    destruct (bbmem_block bbm b).
    specialize (Hbi _ eq_refl).
    apply mem_block_access_max.
    reflexivity.
  Qed.
  Next Obligation.
    unfold ZMap'.get.
    apply bbmem_nextblock_noblock in H.
    rewrite H.
    reflexivity.
  Qed.
  Next Obligation.
    unfold ZMap'.get.
    apply bbmem_nextblock_noblock in H.
    rewrite H.
    reflexivity.
  Qed.
  Next Obligation.
    rewrite bbmem_nextblock_noblock.
    split; destruct (bbmem_block bbm b); (discriminate || tauto).
  Qed.
  Next Obligation.
    case_eq (bbmem_block bbm b).
    * intros blk Hblk.
      simpl.
      apply f_equal.
      eapply bbmem_negblock_default.
      exact H.
      exact Hblk.
    * intro H'.
      pose proof (bbmem_nextblock_noblock b).
      apply proj2 in H0.
      apply H0 in H'; clear H0.
      pose proof bbmem_nextblock_pos.
      omega.
  Qed.
  Next Obligation.
    unfold ZMap'.get.
    apply bbmem_nextblock_noblock in H.
    rewrite H.
    reflexivity.
  Qed.
  Next Obligation.
    unfold ZMap'.get in *.
    destruct (decide (b >= bbmem_nextblock bbm)) as [Hb|Hb].
    * apply bbmem_nextblock_noblock in Hb.
      rewrite Hb.
      reflexivity.
    * rewrite bbmem_nextblock_noblock in Hb.
      pose proof (bbmem_block_invariants b) as Hblk.
      destruct (bbmem_block bbm b) as [blk|].
      - specialize (Hblk blk eq_refl).
        eapply mem_block_bounds_perm; eassumption.
      - reflexivity.
  Qed.

  (** ** This correspondence is an isomorphism. *)

  Lemma mem_bbmem m:
    mem_of_bbmem (bbmem_of_mem m) = m.
  Proof.
    apply mem_ext_eq; simpl;
    try (apply functional_extensionality; intro b;
         unfold get_block;
         destruct (decide (b >= nextblock m)) as [Hb|Hb];
         pose proof Hb as Ht;
         rewrite nextblock_notget in Ht).

    - rewrite Ht; simpl.
      apply functional_extensionality; intro ofs.
      symmetry.
      apply nextblock_nocontents.
      assumption.
    - destruct (tget m b).
      reflexivity.
      elim Ht; reflexivity.

    - rewrite Ht; simpl.
      apply functional_extensionality; intro ofs.
      apply functional_extensionality; intro k.
      symmetry.
      apply nextblock_noaccess.
      assumption.
    - destruct (tget m b).
      reflexivity.
      elim Ht; reflexivity.

    - reflexivity.

    - rewrite Ht; simpl.
      reflexivity.
    - destruct (tget m b).
      reflexivity.
      elim Ht; reflexivity.

    - rewrite Ht; simpl.
      symmetry.
      apply nextblock_nobounds.
      assumption.
    - destruct (tget m b).
      reflexivity.
      elim Ht; reflexivity.
  Qed.

  Lemma bbmem_mem bbm `{Hbbm: bbmem_invariants bbm}:
    bbmem_of_mem (mem_of_bbmem bbm) = bbm.
  Proof.
    apply (bbmem_ext_eq _ _); intro b.
    simpl.
    destruct (decide (b >= bbmem_nextblock bbm)) as [Hb|Hb].
    * rewrite get_block_none by assumption.
      rewrite bbmem_nextblock_noblock in Hb.
      symmetry; assumption.
    * case_eq (get_block (mem_of_bbmem bbm) b). intros blk1 Hblk1.
      case_eq (bbmem_block bbm b). intros blk2 Hblk2.
      apply f_equal.
      apply mem_block_ext_eq.
      - unfold get_block in Hblk1.
        monad_inv Hblk1.
        inversion Hblk1; simpl.
        rewrite Hblk2.
        reflexivity.
      - unfold get_block in Hblk1.
        monad_inv Hblk1.
        inversion Hblk1; simpl.
        rewrite Hblk2.
        reflexivity.
      - unfold get_block in Hblk1.
        monad_inv Hblk1.
        inversion Hblk1; simpl; clear Hblk1 H0.
        unfold mem_of_bbmem in Hblk0; simpl in *.
        monad_inv Hblk0.
        rewrite Hblk1 in Hblk2; clear Hblk1.
        inversion Hblk2; subst; clear Hblk2.
        inversion Hblk0.
        reflexivity.
      - unfold get_block in Hblk1.
        monad_inv Hblk1.
        inversion Hblk1; simpl.
        rewrite Hblk2.
        reflexivity.
      - intro H.
        rewrite <- bbmem_nextblock_noblock in H.
        tauto.
      - intro H.
        replace (bbmem_nextblock bbm) with (nextblock (mem_of_bbmem bbm)) in Hb.
        apply get_block_some in Hb.
        destruct Hb.
        rewrite H0 in H.
        discriminate.
        reflexivity.
  Qed.
End MEMMAP.

(** * Invariants of source/target memory states *)

Section INVARIANTS.
  Import Memimpl Memtype.Mem.

  (** Whether [x] is in the range [ofs, ofs+sz). *)
  Definition in_range (ofs: Z) (sz: Z) (x: Z): Prop :=
    ofs <= x < ofs + sz.

  Instance in_range_dec ofs sz x: Decision (in_range ofs sz x) := {
    decide :=
      match Z_le_dec ofs x with
        | left H1 =>
          match Z_lt_dec x (ofs + sz) with
            | left H2 => left (conj H1 H2)
            | right H2 => right (fun Hr => H2 (proj2 Hr))
          end
        | right H1 => right (fun Hr => H1 (proj1 Hr))
      end
  }.

  (** Whether [ofs, ofs+sz) is entierly included within [bounds = (lo,hi)]. *)

  Definition in_bounds (bounds: Z * Z) (ofs: Z) (sz: Z): Prop :=
    fst bounds <= ofs /\ ofs + sz <= snd bounds.

  Instance in_bounds_dec bounds ofs sz: Decision (in_bounds bounds ofs sz) := {
    decide :=
      match Z_le_dec (fst bounds) ofs with
        | left H1 =>
          match Z_le_dec (ofs + sz) (snd bounds) with
            | left H2 => left (conj H1 H2)
            | right H2 => right (fun Hb => H2 (proj2 Hb))
          end
        | right H1 => right (fun Hb => H1 (proj1 Hb))
      end
  }.

  (** Whether one range is entierly included in another *)

  Definition range_included ofs1 sz1 ofs2 sz2 :=
    ofs1 >= ofs2 /\
    ofs1 + sz1 <= ofs2 + sz2.

  (** Whether two ranges are disjoint. *)

  Definition range_disjoint (ofs1: Z) (sz1: Z) (ofs2: Z) (sz2: Z) :=
    forall (x: Z), ~ (in_range ofs1 sz1 x /\ in_range ofs2 sz2 x).

  (** We expect the source memory to have the following structure:
      - global variables have [Tag_heap] tags and are preserved as-is
        by the injection;
      - stack variables are grouped into "regions" which correspond
        to a given function call.

    Regions are composed of one "region block", tagged with [Stack_main],
    followed by zero or more "variable blocks", tagged with [Stack_var].
      - In the source memory state, the region block is empty;
      - In the target memory state, the contents of all variables are
        mapped into the region block, whereas the variable blocks are
        empty.

    The predicate below asserts that a source memory state is well-formed
    with regards to this structure. *)

  Class IsEmpty (blk: mem_block): Prop :=
    is_empty: mem_block_bounds blk = (0, 0).

  Class IsInMem bbm b blk: Prop :=
    is_in_mem: bbmem_block bbm b = Some blk.

  Class IsHeap blk ht: Prop :=
    is_heap: mem_block_tag blk = Tag_heap ht.

  Class IsRegion blk fi sz: Prop :=
    is_region: mem_block_tag blk = Stack_main fi (Some sz).

  Class IsVariable blk fi vi rb sz ofs al: Prop :=
    is_variable: mem_block_tag blk = Svar fi vi rb sz ofs (Some al).

  Class IsSourceRegion blk fi sz: Prop := {
    source_region_is :> IsRegion blk fi sz;
    source_region_empty :> IsEmpty blk
  }.

  Class IsTargetRegion blk fi sz: Prop := {
    target_region_is :> IsRegion blk fi sz;
    target_region_bounds: mem_block_bounds blk = (0, sz)
  }.

  Class IsRegionVar bbm rb rblk fi rsz vb vblk vi vsz ofs al: Prop := {
    is_region_variable :>
      IsVariable vblk fi vi rb vsz ofs al;
    variable_region_in_mem :>
      IsInMem bbm rb rblk;
    variable_region_region :>
      IsRegion rblk fi rsz;
    variable_in_range:
      range_included ofs vsz 0 rsz;
    variable_in_group b':
      rb < b' < vb ->
      exists blk' fi' vi' sz' ofs' al',
        IsInMem bbm b' blk' /\
        IsVariable blk' fi' vi' rb sz' ofs' al' /\
        range_disjoint ofs vsz ofs' sz'
  }.

  Class IsSourceVar bbm rb rblk fi rsz vb vblk vi vsz ofs al: Prop := {
    is_source_var :>
      IsRegionVar bbm rb rblk fi rsz vb vblk vi vsz ofs al;
    source_var_bounds:
      mem_block_bounds vblk = (0, vsz)
  }.

  Class IsTargetVar bbm rb rblk fi rsz vb vblk vi vsz ofs al: Prop := {
    is_target_var :>
      IsRegionVar bbm rb rblk fi rsz vb vblk vi vsz ofs al;
    target_var_empty :>
      IsEmpty vblk
  }.

  Inductive IsSourceBlock bbm b blk: Prop :=
    | source_block_heap ht:
        IsInMem bbm b blk ->
        IsHeap blk ht ->
        IsSourceBlock bbm b blk
    | source_block_region fi sz:
        IsInMem bbm b blk ->
        IsSourceRegion blk fi sz ->
        IsSourceBlock bbm b blk
    | source_block_var rb rblk fi rsz vi vsz ofs al:
        IsInMem bbm b blk ->
        IsSourceVar bbm rb rblk fi rsz b blk vi vsz ofs al ->
        IsSourceBlock bbm b blk.

  Existing Class IsSourceBlock.
  Global Existing Instances
    source_block_heap
    source_block_region
    source_block_var.

  Inductive IsTargetBlock bbm b blk: Prop :=
    | target_block_heap ht:
        IsInMem bbm b blk ->
        IsHeap blk ht ->
        IsTargetBlock bbm b blk
    | target_block_region fi sz:
        IsInMem bbm b blk ->
        IsTargetRegion blk fi sz ->
        IsTargetBlock bbm b blk
    | target_block_var rb rblk fi rsz vi vsz al ofs:
        IsInMem bbm b blk ->
        IsTargetVar bbm rb rblk fi rsz b blk vi vsz ofs al ->
        IsTargetBlock bbm b blk.

  Existing Class IsTargetBlock.
  Global Existing Instances
    target_block_heap
    target_block_region
    target_block_var.

  Class bbmem_source_invariants (bbm: bbmem): Prop := {
    source_bbmem_invariants :> bbmem_invariants bbm;
    bbmem_source_block_invariants b blk :>
      IsInMem bbm b blk ->
      IsSourceBlock bbm b blk
  }.

  Class bbmem_target_invariants (bbm: bbmem): Prop := {
    target_bbmem_invariants :> bbmem_invariants bbm;
    bbmem_target_block_invariants b blk :>
      IsInMem bbm b blk ->
      IsTargetBlock bbm b blk
  }.
End INVARIANTS.

(** Definition of the alpha function *)

Section ALPHA.
  Import Memtype.Mem.
  Import Memimpl.
  Local Notation "a # b" := (ZMap'.get b a) (at level 1).

  (** Overwriting part of a memory block with a copy of part of another. *)

  Definition perm_map :=
    Z -> perm_kind -> option permission.

  Definition perm_copy dpm dofs spm sofs sz: perm_map :=
    fun ofs k =>
      if (decide (in_range dofs sz ofs)) then
        spm (ofs + sofs - dofs) k
      else
        dpm ofs k.

  Program Definition mem_copy dblk dofs sblk sofs sz :=
    let bytes :=
      getN (Z.to_nat sz) sofs (mem_block_contents sblk) in
    let contents :=
      setN bytes dofs (mem_block_contents dblk) in
    let access :=
      perm_copy (mem_block_access dblk) dofs (mem_block_access sblk) sofs sz in
    {|
      mem_block_tag := mem_block_tag dblk;
      mem_block_contents := contents;
      mem_block_access := access;
      mem_block_bounds := mem_block_bounds dblk
    |}.

  Instance mem_copy_invariants dblk dofs sblk sofs sz:
    mem_block_invariants dblk ->
    mem_block_invariants sblk ->
    in_bounds (mem_block_bounds dblk) dofs sz ->
    in_bounds (mem_block_bounds sblk) sofs sz ->
    mem_block_invariants (mem_copy dblk dofs sblk sofs sz).
  Proof.
    intros Hdblk Hdb Hsb.
    unfold mem_copy, perm_copy in *.
    split; simpl in *.
    * intro ofs.
      destruct (Z_le_dec dofs ofs);
      destruct (Z_lt_dec ofs (dofs + sz));
      apply mem_block_access_max.
    * intros lo hi ofs k Hlh.
      rewrite Hlh in *.
      unfold in_bounds in *; simpl in *.
      destruct (Z_le_dec dofs ofs);
      destruct (Z_lt_dec ofs (dofs + sz));
      try omega;
      apply (mem_block_bounds_perm lo hi ofs k Hlh).
  Qed.

  Import Memtype.Mem.

  (** Build an empty block with the given tag and bounds. *)
  Definition empty_block (t: tag) (lo hi: Z) := {|
    mem_block_tag := t;
    mem_block_bounds := (lo, hi);
    mem_block_contents := fun _ => Undef;
    mem_block_access := fun _ _ => None
  |}.

  Instance empty_block_invariants t lo hi:
    mem_block_invariants (empty_block t lo hi).
  Proof.
    split; reflexivity.
  Qed.

  Instance empty_block_is_empty t:
    IsEmpty (empty_block t 0 0).
  Proof.
    reflexivity.
  Qed.

  Definition blocks_left bbm b := Z.to_nat (bbmem_nextblock bbm - b).

  (** Copy the contents of [n] variable blocks, starting at block [v],
    into the given region [mem_block] *)
  Function copy_vars bbm rb rblk vb {measure (blocks_left bbm) vb}: mem_block :=
    if decide (vb < bbmem_nextblock bbm) then
      match bbmem_block bbm vb with
        | Some vblk =>
            match (mem_block_tag vblk) with
              | Tag_stack (Stack_var (Svar fi vi r sz ofs (Some al))) =>
                  if decide (r = rb) then
                    copy_vars bbm r (mem_copy rblk ofs vblk 0 sz) (vb+1)
                  else
                    rblk
              | _ => rblk
            end
        | None => rblk
      end
    else rblk.
  Proof.
    intros; subst.
    unfold blocks_left.
    apply Z2Nat.inj_lt; omega.
  Qed.

  Lemma is_same_block {bbm b} blk1 blk2
    {Hblk1: IsInMem bbm b blk1}
    {Hblk2: IsInMem bbm b blk2}:
    blk1 = blk2.
  Proof.
    unfold IsInMem in *.
    rewrite Hblk1 in Hblk2.
    inversion Hblk2.
    reflexivity.
  Qed.

  Lemma is_same_region_size {rblk fi1 fi2} sz1 sz2
    {Hr1: IsRegion rblk fi1 sz1}
    {Hr2: IsRegion rblk fi2 sz2}:
    sz1 = sz2.
  Proof.
    unfold IsRegion in *.
    rewrite Hr1 in Hr2.
    inversion Hr2.
    reflexivity.
  Qed.

  Global Instance copy_vars_invariants bbm rb rblk' fi' rsz' rblk:
    bbmem_source_invariants bbm ->
    IsInMem bbm rb rblk' ->
    IsRegion rblk' fi' rsz' ->
    mem_block_bounds rblk = (0, rsz') ->
    mem_block_invariants rblk ->
    mem_block_invariants (copy_vars bbm rb rblk (rb+1)).
  Proof.
    intros Hbbm' Hrblk'1 Hrblk'2.
    functional induction (copy_vars bbm rb rblk (rb+1));
      intros Hrblk1 Hrblk2;
      try tauto.

    (* Only the recursive case remains *)
    apply IHm; clear IHm; try assumption.
    apply mem_copy_invariants; try assumption.
    - apply (bbmem_block_invariants (bbm:=bbm) vb);
      eassumption.
    - rewrite Hrblk1.
      destruct (bbmem_source_block_invariants vb vblk e0).
      * rewrite is_heap in e1.
        discriminate.
      * rewrite is_region in e1.
        discriminate.
      * rewrite is_variable in e1.
        inversion e1; subst; clear e1.
        pose proof (is_same_block rblk0 rblk'); subst.
        pose proof (is_same_region_size rsz rsz'); subst.
        pose proof variable_in_range as Hrange.
        unfold range_included, in_range, in_bounds in *.
        simpl.
        omega.
    - destruct (bbmem_source_block_invariants vb vblk e0).
      * rewrite is_heap in e1.
        discriminate.
      * rewrite is_region in e1.
        discriminate.
      * rewrite is_variable in e1.
        inversion e1; subst; clear e1.
        pose proof (is_same_block rblk0 rblk'); subst.
        pose proof (is_same_region_size rsz rsz'); subst.
        rewrite source_var_bounds.
        split; reflexivity.
  Qed.

  Lemma copy_vars_tag bbm rb rblk vb:
    mem_block_tag (copy_vars bbm rb rblk vb) =
    mem_block_tag rblk.
  Proof.
    functional induction (copy_vars bbm rb rblk vb); try reflexivity.
    rewrite IHm.
    reflexivity.
  Qed.

  Lemma copy_vars_bounds bbm rb rblk vb:
    mem_block_bounds (copy_vars bbm rb rblk vb) =
    mem_block_bounds rblk.
  Proof.
    functional induction (copy_vars bbm rb rblk vb); try reflexivity.
    rewrite IHm.
    reflexivity.
  Qed.

  (** Build a "region" block, by copying the contents of [n] variables,
    starting at block index [v]. *)
  Definition compute_region_block bbm rb fi rsz: mem_block :=
    let blk := empty_block (Stack_main fi (Some rsz)) 0 rsz in
      copy_vars bbm rb blk (rb+1).

  (** Compute block [b] of the target memory state corresponding to the
    source memory state [bbm]. *)
  Definition compute_block bbm b: option mem_block :=
    do blk <- bbmem_block bbm b;
    match mem_block_tag blk with
      | Tag_heap ht =>
          ret blk
      | Tag_stack (Stack_var t) =>
          ret (empty_block t 0 0)
      | Tag_stack (Stack_main fi (Some sz)) =>
          ret (compute_region_block bbm b fi sz)
      | _ =>
          None
    end.

  Lemma compute_block_same_domain (bbm: bbmem) `{bbmem_source_invariants bbm}:
    forall b,
      bbmem_block bbm b = None <->
      compute_block bbm b = None.
  Proof.
    intro b.
    case_eq (bbmem_block bbm b).
    * intros blk Hblk.
      split; try discriminate.
      unfold compute_block.
      rewrite Hblk; simpl.
      destruct (bbmem_source_block_invariants b blk).
      - assumption.
      - rewrite is_heap.
        discriminate.
      - rewrite is_region.
        discriminate.
      - rewrite is_variable.
        discriminate.
    * intros Hblk.
      split; trivial.
      unfold compute_block.
      rewrite Hblk; simpl.
      tauto.
  Qed.

  Instance compute_block_invariants bbm `{bbmem_source_invariants bbm} b blk':
    compute_block bbm b = Some blk' -> mem_block_invariants blk'.
  Proof.
    unfold compute_block.
    case_eq (bbmem_block bbm b); simpl; try discriminate.
    intros blk Hblk.
    destruct (bbmem_source_block_invariants b blk).
    * assumption.
    * rewrite (_: IsHeap blk _).
      intro Hblk'; inv Hblk'.
      eapply bbmem_block_invariants.
      eassumption.
    * rewrite (_: IsRegion blk _ _).
      intro Hblk'; inv Hblk'.
      unfold compute_region_block.
      eapply copy_vars_invariants; try typeclasses eauto.
      reflexivity.
    * rewrite (_: IsVariable blk _ _ _ _ _ _).
      intro Hblk'; inv Hblk'.
      typeclasses eauto.
  Qed.

  Lemma compute_block_tag bbm `{bbmem_source_invariants bbm} b blk blk':
    bbmem_block bbm b = Some blk ->
    compute_block bbm b = Some blk' ->
    mem_block_tag blk' = mem_block_tag blk.
  Proof.
    intros Hblk.
    unfold compute_block.
    rewrite Hblk; simpl.
    induction (bbmem_source_block_invariants b blk).
    * rewrite is_heap.
      intro Hblk'; inv Hblk'.
      apply is_heap.
    * rewrite is_region.
      intro Hblk'; inv Hblk'.
      unfold compute_region_block.
      rewrite copy_vars_tag.
      reflexivity.
    * rewrite is_variable.
      intro Hblk'; inv Hblk'.
      reflexivity.
    * assumption.
  Qed.

  Lemma compute_block_region:
    forall bbm `{Hbbm: bbmem_source_invariants bbm},
    forall b blk `{Hblk: IsInMem bbm b blk},
    forall fi rsz `{Hblk_tag: IsRegion blk fi rsz},
    forall blk',
      compute_block bbm b = Some blk' ->
      IsTargetRegion blk' fi rsz.
  Proof.
    intros ? ? ? ? ? ? ? ? ? Hblk'.
    split.
    * unfold IsRegion.
      erewrite compute_block_tag;
      eassumption.
    * revert Hblk'.
      unfold compute_block.
      rewrite Hblk.
      simpl.
      rewrite Hblk_tag.
      unfold compute_region_block.
      intro H; inv H.
      rewrite copy_vars_bounds.
      reflexivity.
  Qed.

  Lemma compute_block_variable_empty bbm vb vblk vblk' fi vi rb sz al ofs:
    IsInMem bbm vb vblk ->
    IsVariable vblk fi vi rb sz al ofs ->
    compute_block bbm vb = Some vblk' ->
    IsEmpty vblk'.
  Proof.
    intros Hvblk Hisvar.
    unfold compute_block.
    rewrite Hvblk; simpl.
    rewrite Hisvar; simpl.
    intro H; inv H.
    typeclasses eauto.
  Qed.

  Definition bbalpha (bbm: bbmem): bbmem := {|
    bbmem_nextblock := bbmem_nextblock bbm;
    bbmem_block := compute_block bbm
  |}.

  Global Instance bbalpha_invariants bbm:
    bbmem_source_invariants bbm ->
    bbmem_invariants (bbalpha bbm).
  Proof.
    intro Hbbm.
    split; simpl.
    * apply bbmem_nextblock_pos.
    * intro b.
      rewrite bbmem_nextblock_noblock.
      rewrite (compute_block_same_domain bbm).
      reflexivity.
    * intros b blk Hb Hblk.
      case_eq (bbmem_block bbm b).
      intros blk' Hblk'.
      erewrite compute_block_tag; try eassumption.
      eapply bbmem_negblock_default.
      exact Hb.
      exact Hblk'.
      intros H.
      apply compute_block_same_domain in H.
      rewrite H in Hblk.
      discriminate.
      assumption.
    * intros b blk Hblk; inv Hblk.
       apply (compute_block_invariants bbm b).
      assumption.
  Qed.

  Lemma bbalpha_defined_inv bbm b blk':
    forall `{bbmem_source_invariants bbm},
      bbmem_block (bbalpha bbm) b = Some blk' ->
      exists blk, bbmem_block bbm b = Some blk.
  Proof.
    intros Hbbm Hblk'.
    unfold bbalpha in Hblk'; simpl in *.
    pose proof (compute_block_same_domain bbm b).
    destruct (bbmem_block bbm b); eauto.
    rewrite Hblk' in H.
    destruct H.
    specialize (H eq_refl).
    discriminate.
  Qed.

  Lemma bbalpha_region_var bbm rb rblk {fi rsz vb vblk vi vsz ofs al} vblk':
    forall `{Hbbm: bbmem_source_invariants bbm},
    forall `{Hregvar: IsRegionVar bbm rb rblk fi rsz vb vblk vi vsz ofs al},
    forall `{Hvblk: IsInMem bbm vb vblk},
    forall `{Hvblk': IsInMem (bbalpha bbm) vb vblk'},
    exists rblk',
      IsRegionVar (bbalpha bbm) rb rblk' fi rsz vb vblk' vi vsz ofs al.
  Proof.
    intros ? ? ? ?.
    case_eq (bbmem_block (bbalpha bbm) rb).
    * intros rblk' Hrblk'; exists rblk'; simpl in *.
      split; try eassumption.
      - unfold IsVariable.
        erewrite compute_block_tag; try eassumption.
        apply is_variable.
      - unfold IsRegion.
        erewrite (compute_block_tag bbm rb); try eassumption.
        apply is_region.
        apply is_in_mem.
      - apply variable_in_range.
      - intros sb Hsb.
        destruct (variable_in_group sb Hsb)
                 as [sblk [sfi [svi [ssz [sofs [sal]]]]]].
        destruct H as [Hsblk [Hsblk_var Hrange]].
        case_eq (compute_block bbm sb).
        + intros tblk Htblk.
          exists tblk.
          exists sfi.
          exists svi.
          exists ssz.
          exists sofs.
          exists sal.
          split.
          apply Htblk.
          split.
          unfold IsVariable.
          erewrite compute_block_tag; try eassumption.
          eassumption.
        + intro H.
          apply compute_block_same_domain in H; try eassumption.
          rewrite Hsblk in H.
          discriminate.
    * intro H.
      simpl in *.
      apply compute_block_same_domain in H; try eassumption.
      rewrite is_in_mem in H.
      discriminate.
  Qed.

  Global Instance bbalpha_target_invariants bbm:
    bbmem_source_invariants bbm ->
    bbmem_target_invariants (bbalpha bbm) := {}.
  Proof.
    intros b blk' Hblk'.
    destruct (bbalpha_defined_inv bbm b blk' Hblk') as [blk Hblk].
    induction (bbmem_source_block_invariants b blk).
    * eapply target_block_heap.
      - assumption.
      - revert Hblk';
        unfold IsInMem, bbalpha, compute_block; simpl;
        rewrite Hblk; simpl; rewrite H0.
        intro Hblk'; inv Hblk'; eassumption.
    * eapply target_block_region.
      - assumption.
      - eapply compute_block_region; try eassumption.
        typeclasses eauto.
    * edestruct (bbalpha_region_var bbm rb rblk blk') as [vblk' Hvblk'];
        try eassumption.
      eapply target_block_var; try eassumption.
      - split.
        apply Hvblk'.
        eapply compute_block_variable_empty;
          try eassumption.
        typeclasses eauto.
    * eassumption.
  Qed.

  (** Recast [bbalpha] and [bbalpha_inv] in terms of [mem] *)

  Class mem_source_invariants (m: mem): Prop :=
    msi_bb_inv :> bbmem_source_invariants (bbmem_of_mem m).

  Definition alpha (m: mem) `{mem_source_invariants m} :=
    mem_of_bbmem (bbalpha (bbmem_of_mem m)).

  (* Other possibility *)

  Context `{forall m, Decision (mem_source_invariants m)}.

  Definition alpha_opt (m: mem): option mem :=
    do _ <- assert (mem_source_invariants m);
    ret (alpha m).

End ALPHA.

(************************************************************************
                         COMPILES UNTIL HERE
 ************************************************************************)

(*
  (** Rebuild a "variable" block from the corresponding region block *)
  Definition build_variable_block (bbm: bbmem) fi vi r ofs sz al :=
    let vblk := empty_block (Svar fi vi r ofs sz (Some al)) 0 sz in
      do rblk <- bbmem_block bbm r;
      ret (mem_copy vblk 0 rblk ofs (Z.to_nat sz)).

  (** Compute block [b] of the source memory state corresponding to the
    target memory state [bbm]. *)
  Definition compute_block_inv (bbm: bbmem) (b: block): option mem_block :=
    do blk <- bbmem_block bbm b;
    match mem_block_tag blk with
      | Tag_heap ht =>
          ret blk
      | Tag_stack (Stack_main fi (Some sz)) =>
          ret (empty_block (Stack_main fi (Some sz)) 0 0)
      | Tag_stack (Svar fi vi r ofs sz (Some al)) =>
          build_variable_block bbm fi vi r ofs sz al
      | _ =>
          None
    end.
  
  Definition bbalpha_inv (bbm: bbmem): bbmem := {|
    bbmem_nextblock := bbmem_nextblock bbm;
    bbmem_block := compute_block_inv bbm
  |}.

  (** When the block [b] has a [Stack_main] tag in the source, then
    presumably it is an empty "region block", and we need to allocate
    a block large enough to store all of the variables of the region
    in the target. *)

  Definition handle_region (m: mem) (r: block) (f: ident) (sz: Z) (m': mem) :=
    let (m', b) := alloc m' 0 sz (Tag_stack (Stack_main f (Some sz))) in
      do _ <- assert (block_bounds m b = (0, 0));
      do _ <- assert (r = b);
      ret m'.

  Definition handle_region_inv (m': mem) r f sz (m: mem) :=
    let (m, b) := alloc m 0 0 (Tag_stack (Stack_main f (Some sz))) in
      do _ <- assert (block_bounds m' b = (0, sz));
      do _ <- assert (r = b);
      ret m.

  (** Conversely, for [Stack_var] blocks, we need to allocate empty
    blocks in the target, but copy the original contents into the
    corresponding region. *)

  Definition handle_var (m: mem) (b: block) f v r sz al ofs (m': mem) :=
    let (m', b') := alloc m' 0 0 (Svar f v r sz al (Some ofs)) in
    do _ <- assert (b = b');
    mem_copy m' r ofs m b 0 (Z.to_nat sz).

  Definition handle_var_inv (m': mem) (b: block) f v r sz al ofs (m: mem) :=
    let (m, b') := alloc m 0 sz (Svar f v r sz al (Some ofs)) in
    do _ <- assert (b = b');
    mem_copy m b 0 m' r ofs (Z.to_nat sz).

  (** Global variables are simply copied from [m] to [m'] *)

  Definition handle_glob (m: mem) (b: block) (t: htag) (m': mem) :=
    let (lo, hi) := block_bounds m b in
    let (m', b') := alloc m' lo hi (Tag_heap t) in
    do _ <- assert (b = b');
    mem_copy m' b lo m b lo (Z.to_nat (hi - lo)).

  Definition handle_glob_inv (m': mem) (b: block) (t: htag) (m: mem) :=
    handle_glob m' b t m.

  (** Copy the next block [b] into the target memory state [m'] from
    the corresponding block in the source memory state [m]. *)

  Definition handle_block m b m': option Memimpl.mem :=
    do t <- tget m b;
    match t with
      | Tag_stack (Stack_main f (Some sz)) =>
          handle_region m b f sz m'
      | Tag_stack (Svar f v r sz al (Some ofs)) =>
          handle_var m b f v r sz al ofs m'
      | Tag_heap t =>
          handle_glob m b t m'
      | _ =>
          None
    end.

  Definition handle_block_inv m' b m: option Memimpl.mem :=
    do t <- tget m' b;
    match t with
      | Tag_stack (Stack_main f (Some sz)) =>
          handle_region_inv m' b f sz m
      | Tag_stack (Svar f v r sz al (Some ofs)) =>
          handle_var_inv m' b f v r sz al ofs m
      | Tag_heap t =>
          handle_glob_inv m' b t m
      | _ =>
          None
    end.

  (** Iterate [handle_block] to obtain the target memory state corresponding
    to the source memory state [m]. *)

  Fixpoint alpha_rec (m: mem) (m': mem) (b: Z) (n: nat) {struct n} :=
    match n with
      | O =>
          ret m'
      | S n =>
          do m' <- handle_block m b m';
          alpha_rec m m' (b+1) n
    end.

  Fixpoint alpha_rec_inv (m': mem) (m: mem) (b: Z) (n: nat) {struct n} :=
    match n with
      | O => ret m
      | S n =>
          do m <- handle_block_inv m' b m;
          alpha_rec_inv m' m (b+1) n
    end.

  Record alpha_rec_alg (P: mem -> Z -> Prop) (m: mem) := {
    ara_region b f sz m' m1':
      P m' b ->
      tget m b = Some (Tag_stack (Stack_main f (Some sz))) ->
      alloc m' 0 sz (Stack_main f (Some sz)) = (m1', b) ->
      P m1' (b+1);
    ara_var b f v r sz al ofs m' m1' m2':
      P m' b ->
      tget m b = Some (Tag_stack (Svar f v r sz al (Some ofs))) ->
      alloc m' 0 0 (Svar f v r sz al (Some ofs)) = (m1', b) ->
      mem_copy m1' r ofs m b 0 (Z.to_nat sz) = Some m2' ->
      P m2' (b+1);
    ara_glob b t m' lo hi m1' m2':
      P m' b ->
      tget m b = Some (Tag_heap t) ->
      block_bounds m b = (lo, hi) ->
      alloc m' lo hi (Tag_heap t) = (m1', b) ->
      mem_copy m1' b lo m b lo (Z.to_nat (hi - lo)) = Some m2' ->
      P m2' (b+1)
  }.

  Lemma ara_block {P m} (e: alpha_rec_alg P m):
    forall b m' m1',
      P m' b ->
      handle_block m b m' = Some m1' ->
      P m1' (b+1).
  Proof.
    intros b m' m1' Hm'.
    unfold handle_block.
    case_eq (tget m b); try discriminate; simpl; intros t Ht1.
    case_eq t.
    * intros h Hh; subst.
      unfold handle_glob.
      case_eq (block_bounds m b); intros lo hi Hlh.
      case_eq (alloc m' lo hi h); intros m2' b2' Hm2'.
      intros Hm1'; assert_inv Hm1'; subst.
      eapply (ara_glob _ _ e); eassumption.
    * intros s Hs; subst.
      case_eq s.
    - intros f osz Hs; subst.
      case_eq osz; try discriminate.
      intros sz Hsz; subst.
      unfold handle_region.
      case_eq (alloc m' 0 sz (Stack_main f (Some sz))); intros m2' b2' Hm2'.
      intros H; assert_inv H; inversion H; subst.
      eapply (ara_region _ _ e); eassumption.
    - intros [f v rb sz al oofs] Hs; subst.
      case_eq oofs; try discriminate.
      intros ofs Hofs; subst.
      unfold handle_var.
      case_eq (alloc m' 0 0 (Svar f v rb sz al (Some ofs))).
      intros m2' b2' Hm2' H.
      assert_inv H; subst.
      eapply (ara_var _ _ e); eassumption.
  Qed.

  Lemma alpha_rec_ind {m m' b n m1'} (H: alpha_rec m m' b n = Some m1'):
    forall P (e: alpha_rec_alg P m),
      P m' b -> P m1' (b + (Z.of_nat n)).
  Proof.
    intros P e.
    revert m' b m1' H.
    induction n.
    * simpl.
      intros m' b m1' Hm1' e0.
      inversion Hm1'; subst.
      rewrite Z.add_0_r.
      assumption.
    * intros m' b m1'; simpl alpha_rec.
      case_eq (handle_block m b m'); try discriminate.
      intros m2' Hm2' Hm1' e0.
      simpl in Hm1'.
      replace (b + Z.of_nat (S n)) with ((b+1) + Z.of_nat n).
      eapply IHn; try eassumption.
      eapply ara_block; try eassumption.
      rewrite Nat2Z.inj_succ; now ring.
  Qed.

  Definition alpha (m: mem) :=
    alpha_rec m empty 1 (Z.to_nat (nextblock m - 1)).

  Definition alpha_inv (m': mem) :=
    alpha_rec_inv m' empty 1 (Z.to_nat (nextblock m' - 1)).

  Lemma alpha_ind {m m': mem} (H: alpha m = Some m'):
    forall P (e: alpha_rec_alg P m),
      P empty 1 -> P m' (nextblock m).
  Proof.
    intros P e e0.
    unfold alpha in H.
    replace (nextblock m) with (1 + (Z.of_nat (Z.to_nat (nextblock m - 1)))).
    apply (alpha_rec_ind H P e e0).
    rewrite Z2Nat.id.
    - ring.
    - pose proof (nextblock_pos m).
      omega.
  Qed.

  (** From a source memory state [m], extract the corresponding injection. *)

  Definition extract_meminj m b :=
    match (Z_lt_dec b (Mem.nextblock m)) with
      | right _ => None
      | left _ =>
        match tget m b with
          | Some (Tag_stack (Svar f v b' sz al (Some off))) => Some (b', off)
          | _ => Some (b, 0)
        end
    end.
End ALPHA.

Import Memtype.Mem.
Notation mem := Memimpl.mem.

Variable alpha_inv: mem -> option mem.

Hypothesis alpha_inv_alpha:
  forall m m', alpha m = Some m' -> alpha_inv m' = Some m.

Hypothesis alpha_alpha_inv:
  forall m' m, alpha_inv m' = Some m -> alpha m = Some m'.

Inductive inject: meminj -> mem -> mem -> Prop :=
  | inject_intro m m':
      alpha m = Some m' ->
      inject (extract_meminj m) m m'.

Lemma inject_func f1 f2 m m'1 m'2:
  inject f1 m m'1 ->
  inject f2 m m'2 ->
  (f1, m'1) = (f2, m'2).
Proof.
  intros H1 H2.
  destruct H1.
  inversion H2.
  congruence.
Qed.

Lemma inject_inj f1 f2 m1 m2 m':
  inject f1 m1 m' ->
  inject f2 m2 m' ->
  (f1, m1) = (f2, m2).
Proof.
  intros H1 H2.
  destruct H1 as [m1 m' Hm1].
  inversion H2 as [? ? Hm2]; subst; clear H2.
  apply (f_equal (fun x => (extract_meminj x, x))).
  assert (Some m1 = Some m2).
  * rewrite <- (alpha_inv_alpha m1 m') by apply Hm1.
    rewrite <- (alpha_inv_alpha m2 m') by apply Hm2.
    reflexivity.
  * inversion H.
    reflexivity.
Qed.

Lemma inject_other f m m' b r:
  inject f m m' ->
  (* b does not carry a stack tag *)
  (forall st, tget m b <> Some (Tag_stack (Stack_var st))) ->
  (* b is mapped *)
  f b = Some r ->
  (* then it is mapped identically *)
  r = (b, 0).
Proof.
  intros H Htb Hfb.
  destruct H.
  unfold extract_meminj in Hfb.
  destruct (tget m b) as [[h|s]|].
  * destruct (Z_lt_dec b (nextblock m)).
    - inversion Hfb.
      reflexivity.
    - discriminate.
  * destruct s as [f sz | t].
    - destruct (Z_lt_dec b (nextblock m)).
      + inversion Hfb; reflexivity.
      + discriminate.
    - elim (Htb t).
      reflexivity.
  * destruct (Z_lt_dec b (nextblock m)).
    - inversion Hfb; reflexivity.
    - discriminate.
Qed.

(* Stack blocks are pretty much mapped as instructed by their tags *)
Lemma inject_same f m m' b fi vi b' sz al of:
  inject f m m' ->
  tget m b = Some (Tag_stack (Svar fi vi b' sz al (Some of))) ->
  f b = Some (b', of).
Proof.
  intros Hmm' Htb.
  destruct Hmm' as [m].
  unfold extract_meminj.
  destruct (Z_lt_dec b (nextblock m)) as [Hb|Hb].
  - rewrite Htb.
    reflexivity.
  - elim Hb.
    apply tget_valid.
    rewrite Htb.
    discriminate.
Qed.

Lemma tget_inject_same f m m' b fi vi b' sz align ofs:
  inject f m m' ->
  tget m b = Some (Tag_stack (Svar fi vi b' sz align (Some ofs))) ->
  exists f', exists sz',
    tget m' b' = Some (Tag_stack (Stack_main f' (Some sz'))).
Proof.
  (* FIXME: Need well-formedness condition on m, m'. *)
  admit.
Qed.

Lemma tget_inject_var_main_size f m m' b fi vi b' sz al of f' sz':
  inject f m m' ->
  (* [b] is a variable within region [b'] *)
  tget m b = Some (Tag_stack (Svar fi vi b' sz al (Some of))) ->
  (* [b'] is the region *)
  tget m b' = Some (Tag_stack (Stack_main f' (Some sz'))) ->
  (* then b is mapped within the bounds of b' *)
  of + sz <= sz'.
Proof.
  (* FIXME: Need well-formedness condition on m, m'. *)
  admit.
Qed.

Lemma tget_inject_var_size f m m' b fi vi b' sz align ofs:
  inject f m m' ->
  tget m b = Some (Tag_stack (Svar fi vi b' sz align (Some ofs))) ->
  0 <= sz <= Int.max_unsigned  /\ 0 < align /\ (align | sz) /\ 0 <= ofs.
Proof.
  (* FIXME: needs invariants *)
  admit.
Qed.

Lemma tget_inject_main_size f m m' b f' sz':
  inject f m m' ->
  tget m b = Some (Tag_stack (Stack_main f' (Some sz'))) ->
  0 <= sz' <= Int.max_unsigned.
Proof.
  (* FIXME: needs invariants *)
  admit.
Qed.

Lemma tget_handle_region m' b f sz m2':
  handle_region b f sz m' = Some m2' ->
  tget m2' b = Some (Tag_stack (Stack_main f (Some sz))).
Proof.
  unfold handle_region.
  case_eq (alloc m' 0 sz (Tag_stack (Stack_main f (Some sz)))).
  intros m1' b' H' Hm12'.
  destruct (Z.eq_dec b b'); try discriminate.
  inversion Hm12'; subst; clear Hm12'.
  eapply Mem.tget_alloc_same.
  eassumption.
Qed.

Lemma tget_mem_copy m' b' ofs' m b ofs sz m1' bt:
  mem_copy m' b' ofs' m b ofs sz = Some m1' ->
  tget m1' bt = tget m' bt.
Proof.
  unfold mem_copy.
  intros Hm1'.
  assert_inv Hm1'.
  inversion Hm1' as [H]; clear H Hm1'.
  simpl.
  reflexivity.
Qed.

Lemma tget_handle_var m' b m fi vi r sz al ofs m2':
  handle_var m b fi vi r sz al ofs m' = Some m2' ->
  Mem.tget m2' b = Some (Tag_stack (Svar fi vi r sz al (Some ofs))).
Proof.
  unfold handle_var.
  case_eq (alloc m' 0 0 (Svar fi vi r sz al (Some ofs))).
  intros m1' b' Hm1' Hm2'.
  assert_inv Hm2'; subst.
  etransitivity.
  * eapply tget_mem_copy.
    eassumption.
  * eapply tget_alloc_same.
    eassumption.
Qed.

Lemma tget_handle_glob m' b m t m1':
  handle_glob m b t m' = Some m1' ->
  tget m1' b = Some (Tag_heap t).
Proof.
  unfold handle_glob.
  destruct (Memimpl.block_bounds m b).
  case_eq (alloc m' z z0 (Tag_heap t)).
  intros m2' b' Hm2' Hm1'.
  assert_inv Hm1'; subst.
  etransitivity.
  * eapply tget_mem_copy.
    eassumption.
  * eapply tget_alloc_same.
    eassumption.
Qed.

Lemma tget_handle_block m b m' m1':
  handle_block m b m' = Some m1' ->
  tget m1' b = tget m b.
Proof.
  unfold handle_block.
  destruct_tget m b;
  intro H; simpl in H.
  * eapply tget_handle_glob.
    eassumption.
  * eapply tget_handle_region.
    eassumption.
  * eapply tget_handle_var.
    eassumption.
Qed.

Lemma nextblock_inject f m1 m2:
  inject f m1 m2 ->
  nextblock m1 = nextblock m2.
Proof.
  intro H.
  destruct H.
  eapply (alpha_ind H (fun m z => z = nextblock m)).
  split.
  * intros.
    erewrite nextblock_alloc.
    rewrite H0.
    reflexivity.
    eassumption.
  * intros b fi vi r sz al ofs m1' m2' m3' Hb Hbt Hm2' Hm3'.
    unfold mem_copy in Hm3'.
    assert_inv Hm3'.
    inversion Hm3'; clear Hm3' H1; simpl.
    change (Memimpl.nextblock m2') with (nextblock m2').
    erewrite nextblock_alloc.
    rewrite Hb.
    reflexivity.
    eassumption.
  * intros b h m1' lo hi m2' m3' Hb Hbt Hbb Hm2' Hm3'.
    unfold mem_copy in Hm3'.
    assert_inv Hm3'.
    inversion Hm3'; clear Hm3' H1; simpl.
    change (Memimpl.nextblock m2') with (nextblock m2').
    erewrite nextblock_alloc.
    rewrite Hb.
    reflexivity.
    eassumption.
  * rewrite nextblock_empty.
    reflexivity.
Qed.

(* FIXME: *)
Context (tget_pos:
  forall m b, b < 1 -> tget m b = None).

Lemma tget_inject f m1 m2 b:
  inject f m1 m2 ->
  tget m1 b = tget m2 b.
Proof.
  intros H.
  destruct (Z_lt_dec b (nextblock m1)) as [Hb|Hb].
  * destruct H.
    revert Hb.
    apply (alpha_ind H).
    split.
    - intros.
      rewrite !tget_pos.  by assumption.
      reflexivity.
    - intros.
      admit.
    - assumption.
  * pose proof Hb as Hb2.
    rewrite (nextblock_inject f m1 m2 H) in Hb2.
    fold (valid_block m1 b) in Hb.
    fold (valid_block m2 b) in Hb2.
    rewrite tget_valid in Hb.
    rewrite tget_valid in Hb2.
    destruct (tget m1 b).
    exfalso; apply Hb; discriminate.
    destruct (tget m2 b).
    exfalso; apply Hb2; discriminate.
    reflexivity.
Qed.

Lemma mi_freeblocks f m1 m2 b:
  inject f m1 m2 ->
  ~(valid_block m1 b) -> f b = None.
Proof.
  intros Hm Hb.
  destruct Hm.
  unfold extract_meminj.
  destruct (Z_lt_dec b (nextblock m)); tauto.
Qed.

   valid_block_inject_1 f m1 m2 b1 b2 delta:
     f b1 = Some(b2, delta) ->
     inject f m1 m2 ->
     valid_block m1 b1
   ;

   valid_block_inject_2 f m1 m2 b1 b2 delta:
     f b1 = Some(b2, delta) ->
     inject f m1 m2 ->
     valid_block m2 b2
   ;

   perm_inject f m1 m2 b1 b2 delta ofs k p:
     f b1 = Some(b2, delta) ->
     inject f m1 m2 ->
     perm m1 b1 ofs k p -> perm m2 b2 (ofs + delta) k p
   ;

   range_perm_inject:
     forall f m1 m2 b1 b2 delta lo hi k p,
       f b1 = Some(b2, delta) ->
       inject f m1 m2 ->
       range_perm m1 b1 lo hi k p -> range_perm m2 b2 (lo + delta) (hi + delta) k p
   ;

   valid_access_inject f m1 m2 chunk b1 ofs b2 delta p:
     f b1 = Some(b2, delta) ->
     inject f m1 m2 ->     
     (align_chunk chunk | option_tag_align (tget m1 b1)) ->
     valid_access m1 chunk b1 ofs p ->
  valid_access m2 chunk b2 (ofs + delta) p
   ;

valid_pointer_inject f m1 m2 b1 ofs b2 delta:
  f b1 = Some(b2, delta) ->
  inject f m1 m2 ->
  valid_pointer m1 b1 ofs = true ->
  valid_pointer m2 b2 (ofs + delta) = true
   ;

weak_valid_pointer_inject f m1 m2 b1 ofs b2 delta:
  f b1 = Some(b2, delta) ->
  inject f m1 m2 ->
  weak_valid_pointer m1 b1 ofs = true ->
  weak_valid_pointer m2 b2 (ofs + delta) = true
   ;

address_inject f m1 m2 b1 ofs1 b2 delta p:
  inject f m1 m2 ->
  perm m1 b1 (Int.unsigned ofs1) Cur p ->
  f b1 = Some (b2, delta) ->
  Int.unsigned (Int.add ofs1 (Int.repr delta)) = Int.unsigned ofs1 + delta
   ;

valid_pointer_inject_no_overflow f m1 m2 b ofs b' delta:
  inject f m1 m2 ->
  valid_pointer m1 b (Int.unsigned ofs) = true ->
  f b = Some(b', delta) ->
  0 <= Int.unsigned ofs + Int.unsigned (Int.repr delta) <= Int.max_unsigned
   ;

weak_valid_pointer_inject_no_overflow f m1 m2 b ofs b' delta:
  inject f m1 m2 ->
  weak_valid_pointer m1 b (Int.unsigned ofs) = true ->
  f b = Some(b', delta) ->
  0 <= Int.unsigned ofs + Int.unsigned (Int.repr delta) <= Int.max_unsigned
   ;

valid_pointer_inject_val f m1 m2 b ofs b' ofs':
  inject f m1 m2 ->
  valid_pointer m1 b (Int.unsigned ofs) = true ->
  val_inject f (Vptr b ofs) (Vptr b' ofs') ->
  valid_pointer m2 b' (Int.unsigned ofs') = true
   ;

weak_valid_pointer_inject_val f m1 m2 b ofs b' ofs':
  inject f m1 m2 ->
  weak_valid_pointer m1 b (Int.unsigned ofs) = true ->
  val_inject f (Vptr b ofs) (Vptr b' ofs') ->
  weak_valid_pointer m2 b' (Int.unsigned ofs') = true
   ;

inject_no_overlap f m1 m2 b1 b2 b1' b2' delta1 delta2 ofs1 ofs2:
  inject f m1 m2 ->
  b1 <> b2 ->
  f b1 = Some (b1', delta1) ->
  f b2 = Some (b2', delta2) ->
  perm m1 b1 ofs1 Max Nonempty ->
  perm m1 b2 ofs2 Max Nonempty ->
  b1' <> b2' \/ ofs1 + delta1 <> ofs2 + delta2
   ;

different_pointers_inject f m m' b1 ofs1 b2 ofs2 b1' delta1 b2' delta2:
  inject f m m' ->
  b1 <> b2 ->
  valid_pointer m b1 (Int.unsigned ofs1) = true ->
  valid_pointer m b2 (Int.unsigned ofs2) = true ->
  f b1 = Some (b1', delta1) ->
  f b2 = Some (b2', delta2) ->
  b1' <> b2' \/
  Int.unsigned (Int.add ofs1 (Int.repr delta1)) <>
  Int.unsigned (Int.add ofs2 (Int.repr delta2))
   ;

disjoint_or_equal_inject:
  forall f m m' b1 b1' delta1 b2 b2' delta2 ofs1 ofs2 sz,
  inject f m m' ->
  f b1 = Some(b1', delta1) ->
  f b2 = Some(b2', delta2) ->
  range_perm m b1 ofs1 (ofs1 + sz) Max Nonempty ->
  range_perm m b2 ofs2 (ofs2 + sz) Max Nonempty ->
  sz > 0 ->
  b1 <> b2 \/ ofs1 = ofs2 \/ ofs1 + sz <= ofs2 \/ ofs2 + sz <= ofs1 ->
  b1' <> b2' \/ ofs1 + delta1 = ofs2 + delta2
             \/ ofs1 + delta1 + sz <= ofs2 + delta2 
             \/ ofs2 + delta2 + sz <= ofs1 + delta1
   ;

aligned_area_inject:
  forall f m m' b ofs al sz b' delta,
  inject f m m' ->
  al = 1 \/ al = 2 \/ al = 4 \/ al = 8 -> sz > 0 ->
  (al | sz) ->
  range_perm m b ofs (ofs + sz) Cur Nonempty ->
  (al | ofs) ->
  f b = Some(b', delta) ->
  (al | ofs + delta)
   ;

load_inject f m1 m2 chunk b1 ofs b2 delta v1:
  inject f m1 m2 ->
  load chunk m1 b1 ofs = Some v1 ->
  f b1 = Some (b2, delta) ->
  (align_chunk chunk | option_tag_align (tget m1 b1)) ->
  exists v2, load chunk m2 b2 (ofs + delta) = Some v2 /\ val_strong_inject f v1 v2
;

loadv_inject f m1 m2 chunk b1 o1 a2 v1:
  inject f m1 m2 ->
  loadv chunk m1 (Vptr b1 o1) = Some v1 ->
  val_inject f (Vptr b1 o1) a2 ->
  (align_chunk chunk | option_tag_align (tget m1 b1)) ->
  exists v2, loadv chunk m2 a2 = Some v2 /\ val_strong_inject f v1 v2
;

loadbytes_inject f m1 m2 b1 ofs len b2 delta bytes1:
  inject f m1 m2 ->
  loadbytes m1 b1 ofs len = Some bytes1 ->
  f b1 = Some (b2, delta) ->
  exists bytes2, loadbytes m2 b2 (ofs + delta) len = Some bytes2
              /\ list_forall2 (memval_strong_inject f) bytes1 bytes2
;
store_mapped_inject f chunk m1 b1 ofs v1 n1 m2 b2 delta v2:
  inject f m1 m2 ->
  store chunk m1 b1 ofs v1 = Some n1 ->
  f b1 = Some (b2, delta) ->
  val_strong_inject f v1 v2 ->
  (align_chunk chunk | option_tag_align (tget m1 b1)) ->
  exists n2,
    store chunk m2 b2 (ofs + delta) v2 = Some n2
    /\ inject f n1 n2
;
storev_mapped_inject f chunk m1 b1 o1 v1 n1 m2 a2 v2:
  inject f m1 m2 ->
  storev chunk m1 (Vptr b1 o1) v1 = Some n1 ->
  val_inject f (Vptr b1 o1) a2 ->
  val_strong_inject f v1 v2 ->
  (align_chunk chunk | option_tag_align (tget m1 b1)) ->
  exists n2,
    storev chunk m2 a2 v2 = Some n2 /\ inject f n1 n2
;
storebytes_mapped_inject f m1 b1 ofs bytes1 n1 m2 b2 delta bytes2:
  inject f m1 m2 ->
  storebytes m1 b1 ofs bytes1 = Some n1 ->
  f b1 = Some (b2, delta) ->
  list_forall2 (memval_strong_inject f) bytes1 bytes2 ->
  exists n2,
    storebytes m2 b2 (ofs + delta) bytes2 = Some n2
    /\ inject f n1 n2
;
alloc_parallel_inject:
  forall f m1 m2 lo1 hi1 t m1' b1 lo2 hi2,
  inject f m1 m2 ->
  alloc m1 lo1 hi1 t = (m1', b1) ->
  lo2 <= lo1 -> hi1 <= hi2 ->
  forall st, t <> Tag_stack st ->
  exists f', exists m2', exists b2,
  alloc m2 lo2 hi2 t = (m2', b2)
  /\ inject f' m1' m2'
  /\ inject_incr f f'
  /\ f' b1 = Some(b2, 0)
  /\ (forall b, b <> b1 -> f' b = f b)

  (** XXX Modifie *)
;
alloc_region_inject:
  forall f m1 m2,
    inject f m1 m2 ->
    forall sz fi b1,
      alloc m1 0 0 (Tag_stack (Stack_main fi (Some sz))) = (m1, b1) ->
      forall l,
      valid_region_alloc fi l ->
      (forall sz1 al1 of1, In (sz1, al1, of1) l ->
                        (0 <= sz1 /\ 0 < al1 /\ (al1 | sz1) /\ (al1 | of1) /\ (al1 | 8) /\ 0 <= of1 /\ of1 + sz1 <= sz)) ->
      (forall n1 sz1 al1 of1, 
         List.nth_error l n1 = Some (sz1, al1, of1) ->
         forall n2 sz2 al2 of2,
           List.nth_error l n2 = Some (sz2, al2, of2) ->
           n1 <> n2 ->
           (of1 + sz1 <= of2 \/ of2 + sz2 <= of1)) ->
      sz = region_size l ->
      exists f', exists b2,
        alloc m2 0 sz (Tag_stack (Stack_main fi (Some sz))) = (m2, b2) /\
        inject f' (region_alloc b1 m1 l) (alloc_dummyblocks b2 m2 l) /\
        inject_incr f f' /\
        f' b1 = Some (b2, 0) /\
        (forall n sz al of, List.nth_error l n = Some (sz, al, of) ->
                   f' (b1 + 1 + Z.of_nat n) = Some (b2, of)) /\
        forall b, ~ (b1 + 1 <= b < b1 + 1 + Z.of_nat (length l)) ->
                  f' b = f b
;

free_parallel_inject:
  forall f m1 m2,
  inject f m1 m2 ->
  forall b1 b2 o, f b1 = Some (b2, o) ->
                  forall lo hi m1',
                  free m1 b1 lo hi = Some m1' ->
                  exists m2',
                    free m2 b2 (lo + o) (hi + o) = Some m2' /\
                    inject f m1' m2'
;

free_region_inject:
  forall f m1 m2,
    inject f m1 m2 ->
    forall fi l,
      valid_region_alloc fi l ->
      forall b1 b2, 
        f b1 = Some (b2, 0) ->
        forall sz,
          tget m1 b1 = Some (Tag_stack (Stack_main fi (Some sz))) ->
          (forall n sz al of,
             List.nth_error l n = Some (sz, al, of) ->
             tget m1 (b1 + 1 + Z.of_nat n) = Some (Tag_stack (Stack_var b1 sz al (Some of)))) ->
          forall m1',
            region_free (b1 + 1) m1 l = Some m1' ->
            exists m2',
              free m2 b2 0 sz = Some m2' /\
              inject f m1' m2'
;

neutral_inject m:
  inject_neutral (nextblock m) m ->
  inject (flat_inj (nextblock m)) m m
*)
