Require Import Memory.
Require Import Coqlib.

Module Mem.
Export Memtype.Mem.

Section M.
Context `{M: MEM}.

Variable (f: stag -> option stag). (* we cannot change heap tags due to empty mem *)

Function tagf (t: tag): option tag :=
  match t with
    | Tag_heap _ => Some t
    | Tag_stack st =>
      match f st with
        | Some st' => Some (Tag_stack st')
        | _ => None
      end
  end.

Class TAGMAP := {

   tagmap: mem -> option mem
   ;

   tagmap_empty:
     tagmap empty = Some empty
   ;

   tagmap_nextblock:
     forall m m', tagmap m = Some m' ->
       nextblock m' = nextblock m
   ;

   tagmap_load: 
     forall m m', tagmap m = Some m' -> forall chunk b o,
       load chunk m' b o = load chunk m b o
   ;

   tagmap_loadv:
     forall m m', tagmap m = Some m' -> forall chunk v,
       loadv chunk m' v = loadv chunk m v
   ;

   tagmap_loadbytes:
     forall m m', tagmap m = Some m' -> forall b o l,
       loadbytes m' b o l = loadbytes m b o l
   ;

   tagmap_tget_none:
     forall m m', tagmap m = Some m' -> forall b,
       tget m b = None -> tget m' b = None
   ;

   tagmap_tget_some:
     forall m m',
       tagmap m = Some m' ->
       forall t b, tget m b = Some t ->
                   exists t',
                     tagf t = Some t' /\
                     tget m' b = Some t'
   ; 

   tagmap_store:
     forall m chunk b o v m',
       store chunk m b o v =  Some m' ->
       forall tm, tagmap m = Some tm ->
                  exists tm', tagmap m' = Some tm' /\
                    store chunk tm b o v = Some tm'
   ;

   tagmap_storev:
     forall m chunk a v m',
       storev chunk m a v = Some m' ->
       forall tm, tagmap m = Some tm ->
                  exists tm', tagmap m' = Some tm' /\
       storev chunk tm a v = Some tm'
   ;

   tagmap_storebytes:
     forall m b o v m',
       storebytes m b o v =  Some m' ->
       forall tm, tagmap m = Some tm ->
                  exists tm', tagmap m' = Some tm' /\
                    storebytes tm b o v = Some tm'
   ;


   tagmap_alloc:
     forall m lo hi t b m',
       alloc m lo hi t = (m', b) ->
       forall tm, tagmap m = Some tm ->
                  forall t', tagf t = Some t' ->
                               exists tm', tagmap m' = Some tm' /\
                                           alloc tm lo hi t' = (tm', b)
   ;

   tagmap_free:
     forall m b lo hi m',
       free m b lo hi = Some m' ->
       forall tm, tagmap m = Some tm ->
                  exists tm', tagmap m' = Some tm' /\
                              free tm b lo hi = Some tm'
   ;

   tagmap_perm:
     forall m m', tagmap m = Some m' -> forall b o k p,
       perm m' b o k p <-> perm m b o k p
   ;

   tagmap_drop:
     forall m b lo hi p m',
       drop_perm m b lo hi p = Some m' ->
       forall tm, tagmap m = Some tm ->
                  exists tm', tagmap m' = Some tm' /\
                              drop_perm tm b lo hi p = Some tm'
   ;

   tagmap_inj:
     (forall s1 t, f s1 = Some t -> forall s2, f s2 = Some t -> s1 = s2) ->
     forall m1 tm, tagmap m1 = Some tm -> forall m2, tagmap m2 = Some tm -> m1 = m2

}.

Context `{TM: TAGMAP}.

Lemma tagmap_alloc_heap:
     forall m,
       tagmap m = Some m ->
       forall lo hi ht b m',
         alloc m lo hi (Tag_heap ht) = (m', b) ->
         tagmap m' = Some m'
.
Proof.
  intros. 
  exploit tagmap_alloc.
  eassumption.
  eassumption.
  reflexivity.
  destruct 1. destruct H1. congruence.
Qed.

Require Import Globalenvs.

Lemma tagmap_store_zeros: 
  forall m b lo hi m',
    store_zeros m b lo hi = Some m' ->
    forall tm, tagmap m = Some tm ->
               exists tm', tagmap m' = Some tm' /\
    store_zeros tm b lo hi = Some tm'.
Proof.
  intros until hi. functional induction (store_zeros m b lo hi); simpl; try discriminate;
                   intros until tm; rewrite (store_zeros_equation tm);
   rewrite e.
  inv H; eauto.
  intro.
  exploit tagmap_store. eassumption. eassumption. 
  destruct 1. destruct H1. rewrite H2.
  eauto.
Qed.

Lemma tagmap_store_init_data:
  forall F V (ge: _ F V) m b o i m',
    Genv.store_init_data ge m b o i = Some m' ->
    forall tm, tagmap m = Some tm ->
               exists tm', tagmap m' = Some tm' /\
    Genv.store_init_data ge tm b o i = Some tm'.
Proof.
  unfold Genv.store_init_data. destruct i; simpl; eauto using tagmap_store.
  injection 1; intros; subst; eauto.
  destruct (Genv.find_symbol ge i); try discriminate.
  eauto using tagmap_store.
Qed.

Lemma tagmap_store_init_data_list:
forall F V (ge: _ F V) l m b o m',
  Genv.store_init_data_list ge m b o l = Some m' ->
  forall tm, tagmap m = Some tm ->
             exists tm', tagmap m' = Some tm' /\
                         Genv.store_init_data_list ge tm b o l = Some tm'.
Proof.
  induction l; simpl.
   injection 3; intros; subst; eauto.
  intros until m'.
  case_eq (Genv.store_init_data ge m b o a); try discriminate.
  intros.
  exploit tagmap_store_init_data.
  eassumption.
  eassumption.
  destruct 1. destruct H2.
  rewrite H3.
  eauto.
Qed.

Lemma tagmap_alloc_global:
  forall F V (ge: _ F V) m ig m',
    Genv.alloc_global ge m ig = Some m' ->
    forall tm, tagmap m = Some tm ->
               exists tm', tagmap m' = Some tm' /\
                           Genv.alloc_global ge tm ig = Some tm'.
Proof.
  intros until m'.
  unfold Genv.alloc_global.
  destruct ig.
  destruct g.
   case_eq (alloc m 0 1 (Tag_heap Heap_global)).
   intros.
   exploit tagmap_alloc.
   eassumption.
   eassumption.
   reflexivity.
   destruct 1. destruct H2. rewrite H3.
   eauto using tagmap_drop.
  case_eq (alloc m 0 (Genv.init_data_list_size (AST.gvar_init v)) (Tag_heap Heap_global)).
  intros.
  exploit tagmap_alloc.
  eassumption.
  eassumption.
  reflexivity.
  destruct 1. destruct H2.
  rewrite H3; clear H H1 H3. revert H0.
  case_eq (store_zeros m0 b 0 (Genv.init_data_list_size (AST.gvar_init v))); try discriminate.
  intros until 1.
  exploit tagmap_store_zeros.
  eassumption.
  eassumption.
  clear H H2.
  destruct 1. destruct H.
  rewrite H0. clear H0.
  case_eq (Genv.store_init_data_list ge m1 b 0 (AST.gvar_init v)); try discriminate.
  intros until 1.
  exploit tagmap_store_init_data_list.
  eassumption.
  eassumption.
  clear H H0. destruct 1. destruct H. rewrite H0. clear H0.
  eauto using tagmap_drop.
Qed.

Lemma tagmap_alloc_globals:
  forall F V (ge: _ F V) l m m',
    Genv.alloc_globals ge m l = Some m' ->
    forall tm, tagmap m = Some tm -> exists tm', tagmap m' = Some tm' /\
    Genv.alloc_globals ge tm l = Some tm'.
Proof.
  induction l; simpl.
   injection 1; intros; subst; eauto.
  intros until m'.
  case_eq (Genv.alloc_global ge m a); try discriminate.
  intros.
  exploit tagmap_alloc_global.
  eassumption.
  eassumption.
  destruct 1. destruct H2. rewrite H3. eauto.
Qed.

Lemma tagmap_init_mem:
  forall F V (p: _ F V) m,
    Genv.init_mem p = Some m ->
    tagmap m = Some m.
Proof.
  unfold Genv.init_mem. intros.
  exploit tagmap_alloc_globals.
  eassumption.
  eapply tagmap_empty.
  destruct 1. destruct H0. congruence.
Qed.

End M.
End Mem.
